"""RPC client manager that provides facitility to simply connect to the broker."""

import asyncio
import collections.abc
import contextlib
import logging
import time
import typing

from .rpclogin import RpcLogin
from .rpcmessage import RpcMessage
from .rpcparam import shvgett
from .rpcri import rpcri_legacy_subscription
from .rpctransport import RpcClient, connect_rpc_client
from .rpcurl import RpcUrl
from .simplebase import SimpleBase
from .value import SHVType

logger = logging.getLogger(__name__)


class SimpleClient(SimpleBase):
    """SHV client made simple to use.

    You most likely want to use this client instead of using RpcClient directly.

    It is designed as a RPC peer that is connected to the SHV RPC Broker.
    """

    APP_NAME: str = "pyshv-client"

    def __init__(
        self,
        client: RpcClient,
        login: RpcLogin,
        *args: typing.Any,  # noqa ANN401
        reconnects: int = 1,
        **kwargs: typing.Any,  # noqa ANN401
    ) -> None:
        super().__init__(client, *args, **kwargs)
        self.login: RpcLogin = login
        """The login used for logging to the connected broker."""
        self.reconnects: int = reconnects
        """Number of attempted reconnects before giving up.

        Any negative number means no limit in reconnects, zero is for no
        reconnects (client must already be connected) and one for only initial
        connection.

        Reconnects are attepted with increased amount of time between them. This
        increase is capped on cca. 1 minute after six attempts. This doesn't
        include timeouts that are part of the reconnect process itself (such as
        timeout of network socket).

        This affects the real reconnect (when receive would raise
        :class:`EOFError`). The reset is handled automatically no matter this
        settings.
        """
        self._subscribes: set[str] = set()
        self._connected = asyncio.Event()
        self._login_task = asyncio.create_task(self._login())

    # TODO solve type hinting in this method. It seems that right now it is not
    # possible to correctly type hint the args and kwargs to copy Self # params.
    # https://discuss.python.org/t/how-to-properly-hint-a-class-factory-with-paramspec/27941/3
    @classmethod
    async def connect(
        cls: type[typing.Self],
        url: RpcUrl | str,
        *args: typing.Any,  # noqa ANN401
        **kwargs: typing.Any,  # noqa ANN401
    ) -> typing.Self:
        """Connect and login to the SHV broker.

        Any additional parameters after ``url`` are passed to class initialization.

        :param url: SHV RPC URL to the broker
        :return: Connected instance.
        """
        if isinstance(url, str):
            url = RpcUrl.parse(url)
        res = cls(await connect_rpc_client(url), url.login, *args, **kwargs)
        try:
            await res.wait_for_login()
        except Exception as exc:
            await res.disconnect()
            raise exc
        return res

    async def wait_for_login(self) -> None:
        """Wait for login completion.

        This also propagates any exception that we encounter during the login
        process.
        """
        while True:
            try:
                await self._login_task
                return
            except asyncio.CancelledError:
                pass

    async def disconnect(self) -> None:  # noqa: D102
        self.reconnects = 0
        await super().disconnect()

    async def _loop(self) -> None:
        reconnect_attempt = 0
        while self.reconnects < 0 or self.reconnects >= reconnect_attempt:
            if self.client.connected:
                reconnect_attempt = 0
                self._connected.set()
                activity_task = asyncio.create_task(self._activity_loop())
                await super()._loop()
                self.client.disconnect()  # write stays open so close it
                self._connected.clear()
                activity_task.cancel()
                with contextlib.suppress(asyncio.exceptions.CancelledError):
                    await activity_task
            else:
                try:
                    await self.client.reset()
                except Exception as exc:
                    timeout = max(60, 2**reconnect_attempt)
                    logger.info(
                        "%s: Connect failed (waiting %d secs): %s",
                        self.client,
                        timeout,
                        exc,
                    )
                    await asyncio.sleep(timeout)
                else:
                    self.__restart_login()
            reconnect_attempt += 1

    async def _activity_loop(self) -> None:
        """Loop run alongside with :meth:`_loop`. It sends pings to the other side."""
        while True:
            t = time.monotonic() - self.client.last_send
            if t < (self.IDLE_TIMEOUT / 2):
                await asyncio.sleep(self.IDLE_TIMEOUT / 2 - t)
            else:
                await self.ping()

    async def _send(self, msg: RpcMessage) -> None:
        await self._connected.wait()
        if not msg.is_request or msg.path or msg.method not in {"hello", "login"}:
            await self._login_task
        await super()._send(msg)

    def _reset(self) -> None:
        super()._reset()
        self.__restart_login()

    def __restart_login(self) -> None:
        if not self._login_task.done():
            self._login_task.cancel()
        self._login_task = asyncio.create_task(self._login())

    async def _login(self) -> None:
        """Login operation and all steps done to prepare or restore client.

        This includes subscribes restoration after connection reset.

        Be aware when you are overwriting this as this is running in the
        """
        res = await self.call("", "hello")
        nonce = shvgett(res, "nonce", str, "")
        await self.call(
            "",
            "login",
            self.login.to_shv(nonce, {"idleWatchDogTimeOut": int(self.IDLE_TIMEOUT)}),
        )
        # Restore subscriptions
        for ri in self._subscribes:
            await self.__subscribe(ri)

    async def call(  # noqa: D102
        self,
        path: str,
        method: str,
        param: SHVType = None,
        call_attempts: int | None = None,
        call_timeout: float | None = None,
        user_id: str | None = None,
    ) -> SHVType:
        timeout = self.call_timeout if call_timeout is None else call_timeout
        if timeout is not None:
            timeout *= self.call_attempts if call_attempts is None else call_attempts
        async with asyncio.timeout(timeout):
            while True:
                with contextlib.suppress(EOFError):
                    return await super().call(
                        path, method, param, call_attempts, call_timeout, user_id
                    )
                await asyncio.sleep(0)  # Let loop detect disconnect
                await self._connected.wait()

    async def subscribe(self, ri: str) -> bool:
        """Perform subscribe for signals on given path.

        Subscribe is always performed on the node itself as well as all its
        children.

        :param ri: SHV RPC RI for subscription to be added.
        """
        res = await self.__subscribe(ri)
        self._subscribes.add(ri)
        return res

    async def __subscribe(self, ri: str) -> bool:
        compat = not await self.peer_is_shv3()
        return bool(
            await self.call(
                ".broker/currentClient" if not compat else ".broker/app",
                "subscribe",
                ri if not compat else rpcri_legacy_subscription(ri),
            )
        )

    async def unsubscribe(self, ri: str) -> bool:
        """Perform unsubscribe for signals on given path.

        :param ri: SHV RPC RI for subscription to be removed.
        :return: ``True`` in case such subscribe was located and ``False``
          otherwise.
        """
        compat = not await self.peer_is_shv3()
        resp = bool(
            await self.call(
                ".broker/currentClient" if not compat else ".broker/app",
                "unsubscribe",
                ri if not compat else rpcri_legacy_subscription(ri),
            )
        )
        if resp:
            self._subscribes.remove(ri)
        return resp

    def subscriptions(self) -> collections.abc.Iterator[str]:
        """Iterate over all subscriptions.

        Note that this uses local subscription cache. It won't reach for the
        current set of subscriptions on the server. This in not be an issue in
        almost all cases because because servers allow subscription
        modifications only to the current client and you should always use
        :meth:`subscribe` and :meth:`unsubscribe` and in such case there should
        be no way this cache gets invalidated compared to the Broker.
        """
        yield from iter(self._subscribes)



"""erform tests of our examples."""

import pytest

import example_client
from shv import RpcMethodAccess, RpcMethodNotFoundError, shvmeta


@pytest.mark.parametrize(
    "path,method,param,result",
    (
        ("", "ls", "test", True),
        ("test", "ls", None, ["device"]),
        ("test/device", "ls", None, [".app", "track"]),
        ("test/device/track", "ls", None, [str(i) for i in range(1, 9)]),
        ("test/device/track", "ls", "1", True),
        ("test/device/track", "ls", "8", True),
        ("test/device/track", "ls", "0", False),
        ("test/device/track", "ls", "9", False),
        (
            "test/device",
            "dir",
            None,
            [
                {1: "dir", 2: 0, 3: "idir", 4: "odir", 5: RpcMethodAccess.BROWSE},
                {
                    1: "ls",
                    2: 0,
                    3: "ils",
                    4: "ols",
                    5: RpcMethodAccess.BROWSE,
                    6: {"lsmod": "olsmod"},
                },
            ],
        ),
        (
            "test/device/.app",
            "dir",
            None,
            [
                {1: "dir", 2: 0, 3: "idir", 4: "odir", 5: RpcMethodAccess.BROWSE},
                {
                    1: "ls",
                    2: 0,
                    3: "ils",
                    4: "ols",
                    5: RpcMethodAccess.BROWSE,
                    6: {"lsmod": "olsmod"},
                },
                {1: "shvVersionMajor", 2: 2, 4: "Int", 5: RpcMethodAccess.READ},
                {1: "shvVersionMinor", 2: 2, 4: "Int", 5: RpcMethodAccess.READ},
                {1: "name", 2: 2, 4: "String", 5: RpcMethodAccess.READ},
                {1: "version", 2: 2, 4: "String", 5: RpcMethodAccess.READ},
                {1: "date", 2: 2, 4: "DateTime", 5: RpcMethodAccess.READ},
                {1: "ping", 2: 0, 5: RpcMethodAccess.BROWSE},
            ],
        ),
        (
            "test/device/track",
            "dir",
            None,
            [
                {1: "dir", 2: 0, 3: "idir", 4: "odir", 5: RpcMethodAccess.BROWSE},
                {
                    1: "ls",
                    2: 0,
                    3: "ils",
                    4: "ols",
                    5: RpcMethodAccess.BROWSE,
                    6: {"lsmod": "olsmod"},
                },
                {
                    1: "reset",
                    2: 32,
                    5: RpcMethodAccess.COMMAND,
                    63: {"dependencies": "Reset all tracks to their initial state"},
                },
                {
                    1: "lastResetUser",
                    2: 2,
                    3: "Int",
                    4: "StringOrNull",
                    5: RpcMethodAccess.READ,
                },
            ],
        ),
        (
            "test/device/track/1",
            "dir",
            None,
            [
                {1: "dir", 2: 0, 3: "idir", 4: "odir", 5: RpcMethodAccess.BROWSE},
                {
                    1: "ls",
                    2: 0,
                    3: "ils",
                    4: "ols",
                    5: RpcMethodAccess.BROWSE,
                    6: {"lsmod": "olsmod"},
                },
                {
                    1: "get",
                    2: 2,
                    3: "Int",
                    4: "List[Int]",
                    5: RpcMethodAccess.READ,
                    6: {"chng": None},
                    63: {"description": "List of tracks"},
                },
                {
                    1: "set",
                    2: 4,
                    3: "List[Int]",
                    5: RpcMethodAccess.WRITE,
                    63: {"description": "Set track"},
                },
            ],
        ),
        ("test/device/track/4", "dir", "get", True),
        ("test/device/track/4", "dir", "set", True),
        ("test/device/track/4", "dir", "chng", False),
        ("test/device/track/4", "get", None, [0, 1, 2, 3]),
    ),
)
async def test_call(example_device, client, path, method, param, result):
    res = await client.call(path, method, param)
    assert res == result
    assert shvmeta(res) == shvmeta(result)


async def test_set(example_device, client):
    tracks = [3, 2, 1, 0]
    assert await client.call("test/device/track/4", "set", tracks) is None

    res = await client.call("test/device/track/4", "get")
    assert res == tracks
    assert shvmeta(res) == shvmeta(tracks)


async def test_reset(example_device, value_client):
    await value_client.prop_set("test/device/track/2", [42] * 8)
    await value_client.subscribe("test/device/**:*:*")
    assert await value_client.call("test/device/track", "reset") is None
    # Reset should send update on track 2 because it was changed
    assert value_client["test/device/track/2"] == [0, 1]
    # No other notification should be received
    assert len(value_client) == 1
    assert (
        await value_client.call("test/device/track", "lastResetUser")
        == "test:testbroker"
    )


async def test_invalid_request(example_device, client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.call("test/device/track/4", "nosuchmethod")


async def test_no_such_node(example_device, client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.ls("test/device/track/none")
    with pytest.raises(RpcMethodNotFoundError):
        await client.dir("test/device/track/none")


async def test_example_client(shvbroker, url):
    await example_client.example_client(url)



"""Class that provides an easier way to define regular methods and signals."""

from __future__ import annotations

import asyncio
import collections
import collections.abc
import dataclasses

from .rpcmethod import RpcMethodAccess, RpcMethodDesc
from .rpctransport import RpcClient
from .simplebase import SimpleBase
from .value import SHVType

RpcMethodFunc = collections.abc.Callable[
    [str, str, SHVType, RpcMethodAccess, str | None],
    SHVType | asyncio.Future[SHVType],
]


class SimpleMethods(SimpleBase):
    """SHV RPC methods and signals implementation helper.

    The regular way to implement methods in :class:`SimpleBase` is by defining
    `_ls`, `_dir`, and `_method_call` methods. That provides ability to
    implement discoverability as well as method implementation in various
    dynamic ways. But in real world applications it is common that we just want
    to define custom methods and having to implement these methods can be error
    prune and unnecessary complex. This class instead provides a way to define
    methods that are discovered and their integration is handled automatically.
    """

    def __init__(
        self,
        client: RpcClient,
        *args: typing.Any,  # noqa ANN401
        **kwargs: typing.Any,  # noqa ANN401
    ) -> None:
        super().__init__(client, *args, **kwargs)
        self.methods: dict[str, dict[str, SimpleMethods.Method]] = (
            collections.defaultdict(dict)
        )
        for attr in self.__dict__.values():
            match attr:
                case self.Method():
                    # TODO check if method doesn't already exist
                    self.methods[attr.path][attr.desc.name] = attr
                case self.Property():
                    self.methods[attr.path].update(attr.methods())

    async def _method_call(
        self,
        path: str,
        method: str,
        param: SHVType,
        access: RpcMethodAccess,
        user_id: str | None,
    ) -> SHVType:
        if path in self.methods:
            if method in self.methods[path]:
                res = self.methods[path][method].func(
                    path, method, param, access, user_id
                )
                if isinstance(res, asyncio.Future):
                    res = await res
                return res
        return await super()._method_call(path, method, param, access, user_id)

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        yield from super()._ls(path)
        for pth in self.methods:
            if not path or pth.startswith(path + "/"):
                yield pth[len(path) :].partition("/")[0]

    def _dir(self, path: str) -> collections.abc.Iterator[RpcMethodDesc]:
        yield from super()._dir(path)
        if path in self.methods:
            for method in self.methods[path].values():
                yield method.desc

    @dataclasses.dataclass
    class Method:
        """The definition of the RPC Method.

        This allows you to define methods as attributes of the object and such
        methods are automatically discovered and used.
        """

        path: str
        """
        """
        desc: RpcMethodDesc
        """
        """
        func: RpcMethodFunc
        """
        """

        def __get__(
            self, instance: SimpleMethods, owner: object
        ) -> collections.abc.Callable:
            return lambda *args, **kwargs: self(instance, *args, **kwargs)

        async def __call__(
            self,
            simple_methods: SimpleMethods,
            signal: str | None = None,
            value: SHVType = None,
            access: RpcMethodAccess = RpcMethodAccess.READ,
        ) -> None:
            """External call is used to send signals."""
            if signal is None:
                if not self.desc.signals:
                    raise NotImplementedError("Method doesn't have any signals")
                signal = next(iter(self.desc.signals.keys()))
            elif signal not in self.desc.signals:
                raise ValueError(f"Invalid signal name: {signal}")
            await simple_methods._signal(
                self.path, signal, self.desc.name, value, access
            )

    @classmethod
    def rpcmethod(
        cls, path: str, desc: RpcMethodDesc
    ) -> collections.abc.Callable[[RpcMethodFunc], SimpleBase.Method]:
        """Decorate method to turn it to :class:`SimpleMethods.Method`."""

        def decorator(func: RpcMethodFunc) -> SimpleBase.Method:
            return cls.Method(path, desc, func)

        return decorator

    @dataclasses.dataclass
    class Property:
        """The definition of the RPC Property node.

        This allows you to define methods as attributes of the object and such
        methods are automatically discovered and used.
        """

        path: str
        """
        """
        tp: str = "Any"
        """
        """
        access: RpcMethodAccess = RpcMethodAccess.READ
        """
        """
        signal: bool | str = False

        def setter(self) -> SimpleMethods.Property:
            # TODO decorate set method
            return self

        def __get__(
            self, instance: SimpleMethods, owner: object
        ) -> collections.abc.Callable:
            return lambda *args, **kwargs: self(instance, *args, **kwargs)

        async def __call__(
            self,
            simple_methods: SimpleMethods,
            value: SHVType = None,
        ) -> None:
            """External call is used to send signal."""
            if not self.signal:
                raise NotImplementedError("Method doesn't have any signals")
            await simple_methods._signal(
                self.path,
                "get",
                "chng" if isinstance(self.signal, bool) else self.signal,
                value,
                self.access,
            )

        def methods(self) -> dict[str, SimpleMethods.Method]:
            """Get methods for property implementation."""
            return {
                "get": SimpleMethods.Method(
                    self.path, RpcMethodDesc.getter(), self._get
                ),
                "set": SimpleMethods.Method(
                    self.path, RpcMethodDesc.setter(), self._set
                ),
            }

        def _get(
            self,
            path: str,
            method: str,
            param: SHVType,
            access: RpcMethodAccess,
            user_id: str | None,
        ) -> SHVType:
            return None

        def _set(
            self,
            path: str,
            method: str,
            param: SHVType,
            access: RpcMethodAccess,
            user_id: str | None,
        ) -> SHVType:
            return None



"""Connection over Unix domain named socket."""

import asyncio
import collections.abc
import logging

from .abc import RpcClient
from .stream import (
    RpcClientStream,
    RpcProtocolSerial,
    RpcServerStream,
    RpcTransportProtocol,
)

logger = logging.getLogger(__name__)


class RpcClientUnix(RpcClientStream):
    """RPC connection to some SHV peer over Unix domain named socket."""

    def __init__(
        self,
        location: str = "shv.sock",
        protocol: type[RpcTransportProtocol] = RpcProtocolSerial,
    ) -> None:
        super().__init__(protocol)
        self.location = location

    def __str__(self) -> str:
        return f"unix:{self.location}"

    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await asyncio.open_unix_connection(self.location)


class RpcServerUnix(RpcServerStream):
    """RPC server listenting for SHV connections on Unix domain named socket."""

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], None | collections.abc.Awaitable[None]
        ],
        location: str = "shv.sock",
        protocol: type[RpcTransportProtocol] = RpcProtocolSerial,
    ) -> None:
        super().__init__(client_connected_cb, protocol)
        self.location = location

    def __str__(self) -> str:
        return f"server.unix:{self.location}"

    async def _create_server(self) -> asyncio.Server:
        return await asyncio.start_unix_server(self._client_connect, path=self.location)

    async def listen(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        await super().listen()
        if not was_listening:
            logger.debug("%s: Listening for clients", self)

    def close(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        super().close()
        if was_listening and (self._server is None or self._server.is_serving()):
            logger.debug("%s: No longer listening for clients", self)

    async def _client_connect(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        logger.debug("%s: New client %s", self, writer.get_extra_info("peername"))
        await super()._client_connect(reader, writer)

    class Client(RpcServerStream.Client):
        """RPC client for Unix server connection."""

        def __str__(self) -> str:
            return f"unix:{self._writer.get_extra_info('peername')}"



"""Implementation of CPON ChainPack conversion tool."""

import argparse
import functools
import sys

from .. import cpon
from ..__version__ import VERSION
from .cpconv import CPFormat, convert


def parse_args() -> argparse.Namespace:
    """Parse passed arguments and return result."""
    parser = argparse.ArgumentParser(
        "pycp2cp2", description="Chainpack to Cpon and vice versa converter"
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
    )
    igroup = parser.add_mutually_exclusive_group()
    ogroup = parser.add_mutually_exclusive_group()
    igroup.add_argument(
        "--ic",
        "--input-chainpack",
        action="store_true",
        help="Input is in Chainpack format (default if no input set).",
    )
    ogroup.add_argument(
        "--oc",
        "--output-chainpack",
        action="store_true",
        help="Output is in Chainpack format.",
    )
    igroup.add_argument(
        "--ip",
        "--input-cpon",
        action="store_true",
        help="Input is in Cpon format.",
    )
    ogroup.add_argument(
        "--op",
        "--output-cpon",
        action="store_true",
        help="Output is in Cpon format (default if no output set).",
    )
    parser.add_argument(
        "-i",
        "--indent-cpon",
        action="store",
        default="",
        help='Indent used when generating Cpon (default is no-indent "")',
    )
    parser.add_argument(
        "--o",
        "--output",
        default="-",
        type=argparse.FileType("wb"),
        help="Output file (stdout is used if not specified).",
    )
    parser.add_argument(
        "INPUT",
        type=argparse.FileType("rb"),
        help="Input file for conversion.",
    )
    return parser.parse_args()


def main() -> int:
    """Application's entrypoint."""
    args = parse_args()

    inf = CPFormat.CPON if not args.ic else CPFormat.CHAINPACK
    outf = CPFormat.CPON if not args.oc else CPFormat.CHAINPACK

    indent_cpon = functools.reduce(
        lambda s, repl: s.replace(*repl),
        (("\\t", "\t"), ("\\r", "\r"), ("\\n", "\n")),
        args.indent_cpon,
    )
    cpon_options = cpon.CponWriter.Options(indent=indent_cpon)

    convert(args.INPUT, inf, args.o, outf, cpon_options)
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Cpon data format reader and writer."""

import collections.abc
import dataclasses
import datetime
import decimal
import io
import typing

from . import commonpack
from .value import (
    SHVIMapType,
    SHVListType,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVType,
    SHVUInt,
)


class Cpon:
    """Cpon constans and definitions."""

    ProtocolType = 2

    @classmethod
    def unpack(cls, data: bytes | str) -> SHVType:
        """Unpack single value from given data."""
        return CponReader.unpack(data)

    @classmethod
    def pack(
        cls, value: SHVType, options: typing.Optional["CponWriter.Options"] = None
    ) -> str:
        """Pack given value and return string."""
        stream = io.BytesIO()
        self = CponWriter(stream, options)
        self.write(value)
        return stream.getvalue().decode("utf-8")


class CponReader(commonpack.CommonReader):
    """Read data in Cpon format."""

    def _skip_white_insignificant(self) -> None:
        while True:
            b = self._peek_byte()
            if b == 0:
                return
            if b > ord(" "):
                if b == ord("/"):
                    self._peek_drop()
                    b = self._read_byte()
                    if b == ord("*"):
                        # multiline_comment_entered
                        while True:
                            b = self._read_byte()
                            if b == ord("*"):
                                b = self._read_byte()
                                if b == ord("/"):
                                    break
                    elif b == ord("/"):
                        # to end of line comment entered
                        while True:
                            b = self._read_byte()
                            if b == ord("\n"):
                                break
                    else:
                        raise ValueError("Malformed comment")
                elif b in {ord(":"), ord(",")}:
                    self._peek_drop()
                    continue
                else:
                    break
            else:
                self._peek_drop()

    def read_meta(self) -> SHVMetaType | None:  # noqa: D102
        self._skip_white_insignificant()
        b = self._peek_byte()
        if b != ord("<"):
            return None
        return self._read_map(">")

    def read(self) -> SHVType:  # noqa: D102
        meta = self.read_meta()

        value: SHVType
        self._skip_white_insignificant()
        b = self._peek_byte()
        if ord("0") <= b <= ord("9") or b == ord("+") or b == ord("-"):
            value = self._new_read_number()
        elif b == ord('"'):
            value = self._read_cstring()
        elif b == ord("["):
            value = self._read_list()
        elif b == ord("{"):
            value = typing.cast(SHVMapType, self._read_map())
        elif b == ord("i"):
            self._peek_drop()
            b = self._peek_byte()
            if b != ord("{"):
                raise ValueError("Invalid IMap prefix.")
            value = typing.cast(SHVIMapType, self._read_map())
        elif b == ord("d"):
            self._peek_drop()
            b = self._peek_byte()
            if b != ord('"'):
                raise ValueError("Invalid DateTime prefix.")
            value = self._read_datetime()
        elif b == ord("b"):
            self._peek_drop()
            b = self._peek_byte()
            if b != ord('"'):
                raise ValueError("Invalid Blob prefix.")
            value = self._read_blob()
        elif b == ord("x"):
            self._peek_drop()
            b = self._peek_byte()
            if b != ord('"'):
                raise ValueError("Invalid HexBlob prefix.")
            value = self._read_hexblob()
        elif b == ord("t"):
            self._read_check(b"true")
            value = True
        elif b == ord("f"):
            self._read_check(b"false")
            value = False
        elif b == ord("n"):
            self._read_check(b"null")
            value = None
        else:
            raise ValueError("Malformed Cpon input.")
        if meta is not None:
            value = SHVMeta.new(value, meta)
        return value

    def _read_datetime(self) -> datetime.datetime:
        self._peek_drop()  # eat '"'
        date = ""
        while True:
            c = chr(self._read_byte())
            if c == '"':
                break
            date += c
        return datetime.datetime.fromisoformat(date)

    @staticmethod
    def _hexdigit_to_int(b: int) -> int:
        if ord("0") <= b <= ord("9"):
            val = b - 48
        elif ord("a") <= b <= ord("f"):
            val = b - ord("a") + 10
        elif ord("A") <= b <= ord("F"):
            val = b - ord("A") + 10
        else:
            raise ValueError(f"Invalid HEX digit: {chr(b)}")
        return val

    def _read_blob(self) -> bytes:
        self._peek_drop()
        res = bytearray()
        while True:
            b = self._read_byte()
            if b == ord("\\"):
                b = self._read_byte()
                if b == ord("\\"):
                    res += b"\\"
                elif b == ord('"'):
                    res += b'"'
                elif b == ord("n"):
                    res += b"\n"
                elif b == ord("r"):
                    res += b"\r"
                elif b == ord("t"):
                    res += b"\t"
                else:
                    hi = b
                    lo = self._read_byte()
                    res.append(
                        16 * CponReader._hexdigit_to_int(hi)
                        + CponReader._hexdigit_to_int(lo)
                    )
            else:
                if b == ord('"'):
                    break  # end of string
                res.append(b)
        return bytes(res)

    def _read_hexblob(self) -> bytes:
        res = bytearray()
        self._peek_drop()
        while True:
            b = self._read_byte()
            if b == ord('"'):
                # end of string
                break
            hi = b
            lo = self._read_byte()
            res.append(
                16 * CponReader._hexdigit_to_int(hi) + CponReader._hexdigit_to_int(lo)
            )
        return bytes(res)

    def _read_cstring(self) -> str:
        res = bytearray()
        self._peek_drop()  # eat '"'
        while True:
            b = self._read_byte()
            if b == ord("\\"):
                b = self._read_byte()
                if b == ord("\\"):
                    res += b"\\"
                elif b == ord("b"):
                    res += b"\b"
                elif b == ord('"'):
                    res += b'"'
                elif b == ord("f"):
                    res += b"\f"
                elif b == ord("n"):
                    res += b"\n"
                elif b == ord("r"):
                    res += b"\r"
                elif b == ord("t"):
                    res += b"\t"
                elif b == ord("0"):
                    res += b"\0"
                else:
                    res += bytes((b,))
            else:
                if b == ord('"'):
                    break  # end of string
                res += bytes((b,))
        # Note: this is required to correctly decode the UTF-8 string that contains
        # multiple bytes.
        return res.decode("utf-8")

    def _read_list(self) -> SHVListType:
        res = []
        self._peek_drop()
        while True:
            self._skip_white_insignificant()
            b = self._peek_byte()
            if b == ord("]"):
                self._peek_drop()
                break
            res.append(self.read())
        return res

    def _read_map(self, terminator: str = "}") -> dict[str | int, SHVType]:
        res: dict[str | int, SHVType] = {}
        self._peek_drop()
        while True:
            self._skip_white_insignificant()
            b = self._peek_byte()
            if b == ord(terminator):
                self._peek_drop()
                break
            key = self.read()
            if not isinstance(key, str | int):
                raise ValueError(f"Invalid Map key: {type(key)}")
            self._skip_white_insignificant()
            val = self.read()
            res[key] = val
        return res

    def _new_read_number(self) -> int | SHVUInt | decimal.Decimal:
        bval = bytearray()

        def accept(possib: bytes) -> int | None:
            b = self._peek_byte()
            if b not in possib:
                return None
            bval.append(self._read_byte())
            return b

        tp: type = int
        parsed = False
        nonempty = False
        accept(b"-+")
        if accept(b"0") is not None:
            if accept(b"x") is not None:
                while accept(b"0123456789AaBbCcDdEeFf") is not None:
                    nonempty = True
                parsed = True
            elif accept(b"b"):
                while accept(b"01"):
                    nonempty = True
                parsed = True
        if not parsed:
            exp = False
            while True:
                lb = accept(
                    b"0123456789.ueE"
                    if tp is int
                    else b"0123456789"
                    if exp
                    else b"0123456789eE"
                )
                if lb == ord("."):
                    tp = decimal.Decimal
                elif lb in {ord("e"), ord("E")}:
                    tp = decimal.Decimal
                    accept(b"-+")
                    exp = True
                elif lb == ord("u"):
                    tp = SHVUInt
                    bval.pop()  # remove "u" from bytes
                    break
                elif lb is None:
                    break
                nonempty = True
        if not nonempty:
            bval.append(ord("0"))

        if tp in {int, SHVUInt}:
            return tp(bval, 0)  # type: ignore
        if tp == decimal.Decimal:
            return tp(bval.decode("ascii"))  # type: ignore
        assert False  # should be unreachable


class CponWriter(commonpack.CommonWriter):
    """Write data in Cpon format."""

    @dataclasses.dataclass
    class Options:
        """Options for the CponWriter."""

        indent: bytes = b""
        """Bytes or string used to indent the code."""

    def __init__(
        self, stream: typing.IO | None = None, options: Options | None = None
    ) -> None:
        super().__init__(stream)
        self.options = options if options is not None else self.Options()
        self._nest_level = 0

    def _indent_item(self, is_online_container: bool, item_index: int) -> None:
        if not self.options.indent:
            return
        if is_online_container:
            if item_index > 0:
                self._writestr(" ")
        else:
            self._writestr("\n")
            self._write(self.options.indent * self._nest_level)

    def write_meta(self, meta: SHVMetaType) -> None:  # noqa: D102
        self._writestr("<")
        self._write_map_content(meta)
        self._writestr(">")

    def write_null(self) -> None:  # noqa: D102
        self._writestr("null")

    def write_bool(self, value: bool) -> None:  # noqa: D102
        self._writestr("true" if value else "false")

    def write_blob(self, value: bytes | bytearray) -> None:  # noqa: D102
        self._writestr('b"')
        for d in value:
            if d >= 0x7F:
                self._writestr("\\")
                self._writestr(self._nibble_to_hexdigit(d // 16))
                self._writestr(self._nibble_to_hexdigit(d % 16))
            else:
                self._writestr(
                    {
                        ord("\0"): "\\0",
                        ord("\\"): "\\\\",
                        ord("\t"): "\\t",
                        ord("\b"): "\\b",
                        ord("\r"): "\\r",
                        ord("\n"): "\\n",
                        ord('"'): '\\"',
                    }.get(d, chr(d))
                )
        self._writestr('"')

    @staticmethod
    def _nibble_to_hexdigit(b: int) -> str:
        if b >= 0:
            if b < 10:
                return chr(b + ord("0"))
            if b < 16:
                return chr(b - 10 + ord("a"))
        raise ValueError(f"Invalid nibble value: {b}")

    def write_string(self, value: str) -> None:  # noqa: D102
        self.write_cstring(value)

    def write_cstring(self, value: str) -> None:  # noqa: D102
        self._writestr('"')
        for d in value:
            self._writestr(
                {
                    "\0": "\\0",
                    "\\": "\\\\",
                    "\t": "\\t",
                    "\b": "\\b",
                    "\r": "\\r",
                    "\n": "\\n",
                    '"': '\\"',
                }.get(d, d)
            )
        self._writestr('"')

    def write_int(self, value: int) -> None:  # noqa: D102
        self._writestr(str(value))

    def write_uint(self, value: int) -> None:  # noqa: D102
        self._writestr(f"{value}u")

    def write_double(self, value: float) -> None:  # noqa: D102
        self._writestr(str(value))

    def write_decimal(self, value: decimal.Decimal) -> None:  # noqa: D102
        sval = str(value)
        if not any(c in ".eE" for c in sval):
            sval += ".0"
        self._writestr(sval)

    def write_datetime(self, value: datetime.datetime) -> None:  # noqa: D102
        # We perform here some modification to make the format more compact in
        # some cases. This is not essentailly required but makes it more
        # compatible with other implementations.
        self._writestr('d"')
        rstr = value.isoformat(
            timespec="milliseconds" if value.microsecond else "seconds"
        )
        rlen = 23 if value.microsecond else 19
        self._writestr(rstr[:rlen])
        rstr = rstr[rlen:]
        if rstr[:6] == "+00:00":
            self._writestr("Z")
        elif rstr[:1] == "Z":
            self._writestr(rstr)
        elif rstr[:1] == "+" or rstr[:1] == "-":
            self._writestr(rstr[:3])
            if (mpart := rstr[4:6]) != "00":
                self._writestr(mpart)
        self._writestr('"')

    def write_list(self, value: SHVListType) -> None:  # noqa: D102
        self._nest_level += 1
        is_oneliner = self._is_oneline_list(value)
        self._writestr("[")
        for i, v in enumerate(value):
            if i > 0:
                self._writestr(",")
            self._indent_item(is_oneliner, i)
            self.write(v)
        self._nest_level -= 1
        self._indent_item(is_oneliner, 0)
        self._writestr("]")

    @staticmethod
    def _is_oneline_list(lst: collections.abc.Sequence[SHVType]) -> bool:
        return len(lst) <= 10 and all(not isinstance(v, list | dict) for v in lst)

    def write_imap(self, value: SHVIMapType) -> None:  # noqa: D102
        self._writestr("i{")
        self._write_map_content(value)
        self._writestr("}")

    def write_map(self, value: SHVMapType) -> None:  # noqa: D102
        self._writestr("{")
        self._write_map_content(value)
        self._writestr("}")

    @staticmethod
    def _is_oneline_map(mmap: collections.abc.Mapping) -> bool:
        return len(mmap) <= 10 and all(
            not isinstance(v, list | dict) for v in mmap.values()
        )

    def _write_map_content(self, mmap: collections.abc.Mapping) -> None:
        self._nest_level += 1
        is_oneliner = self._is_oneline_map(mmap)
        i = -1
        for ikey in sorted(k for k in mmap.keys() if isinstance(k, int)):
            i += 1
            if i > 0:
                self._writestr(",")
            self._indent_item(is_oneliner, i)
            self.write_int(ikey)
            self._writestr(":")
            self.write(mmap[ikey])
        for skey in sorted(k for k in mmap.keys() if isinstance(k, str)):
            i += 1
            if i > 0:
                self._writestr(",")
            self._indent_item(is_oneliner, i)
            self._writestr('"')
            self._writestr(skey)
            self._writestr('":')
            self.write(mmap[skey])
        self._nest_level -= 1
        self._indent_item(is_oneliner, 0)



"""Common extension for the SimpleClient."""

import asyncio
import collections.abc
import datetime
import logging
import time
import typing

from .rpcerrors import RpcMethodCallExceptionError, RpcMethodNotFoundError
from .rpcri import rpcri_match
from .simpleclient import SimpleClient
from .value import SHVMapType, SHVType, shvmeta, shvmeta_eq

logger = logging.getLogger(__name__)


class ValueClient(SimpleClient, collections.abc.Mapping):
    """SHV client made to track values more easily.

    This tailors to the use case of tracking and accessing various values more
    easily. You need to subscribe to specific path and this class automatically
    provides you with cached latest value as received through signals or fetched
    from logs (logs fetching has to be performed explicitly) or with prop_get.

    To access subscribed value you can index this object with SHV path to it.
    """

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa ANNN401
        super().__init__(*args, **kwargs)  # notype
        self._cache: dict[str, tuple[float, SHVType]] = {}
        self._handlers: dict[
            str, collections.abc.Callable[[ValueClient, str, SHVType], None]
        ] = {}
        self._futures: dict[str, list[asyncio.Future]] = {}

    def __getitem__(self, key: str) -> SHVType:
        return self._cache[key][1]

    def __iter__(self) -> collections.abc.Iterator[str]:
        return iter(self._cache.keys())

    def __len__(self) -> int:
        return len(self._cache)

    async def _value_update(self, path: str, value: SHVType) -> None:
        handler = self._get_handler(path)
        if handler is not None and not shvmeta_eq(self.get(path, None), value):
            handler(self, path, value)
        for future in self._futures.pop(path, []):
            future.set_result(value)
        if self.is_subscribed(path):
            # Theoretically we should get only paths we subscribed for but user might
            # invoke subscribe on its own which could break our cache logic and thus
            # just guard against it here.
            self._cache[path] = (time.time(), value)

    def _get_handler(
        self, path: str
    ) -> None | collections.abc.Callable[["ValueClient", str, SHVType], None]:
        """Get the handler for the longest path match."""
        split_key = path.split("/")
        paths = (
            "/".join(split_key[: len(split_key) - i]) for i in range(len(split_key) + 1)
        )
        return next(
            (self._handlers[path] for path in paths if path in self._handlers),
            None,
        )

    async def prop_get(self, path: str, max_age: float = 0.0) -> SHVType:
        """Get value from the property associated with the node on given path.

        This always calls get method compared to item access that is served only from
        cache.

        :param path: SHV path to the property node.
        :param max_age: is maximum age in seconds to be specified for the get. Nonzero
          value results in value to be served from cache anywhere along the way (thus not
          just local cache).
        :return: Value of the property node.
        """
        # Serve from cache if cache was updated not before max_age
        if path in self._cache and self._cache[path][0] + max_age >= time.time():
            return self._cache[path][1]
        value = await self.call(path, "get", max_age if int(max_age * 1000) else None)
        if self.get(path, max_age if max_age else None) != value:
            await self._value_update(path, value)
        return value

    async def prop_set(self, path: str, value: SHVType, update: bool = False) -> None:
        """Set value to the property associated with the node on given path.

        :param path: SHV path to the property node.
        :param value: Value to be set to the property node.
        :param update: If internal cache should be immediatelly updated. Otherwise you
            cache would be updated only with get or change signal.
        """
        await self.call(path, "set", value)
        if update:
            await self._value_update(path, value)

    async def prop_change_wait(
        self,
        path: str,
        value: SHVType = None,
        timeout: float | int | None = 5.0,  # noqa ASYNC109
        get_period: float = 1.0,
    ) -> SHVType:
        """Wait for property change.

        The wait is implemented by combination of pooling and waiting for the change
        signal. Pooling uses :meth:`prop_get` every ``get_period``. If you previously
        subscribed on this path (that includes its parent) then it will wait for signal,
        otherwise it will just sleep between get attempts.

        note:: This uses :meth:`on_change` internally and thus it might temporally (for
        this method execution time) replace your own callback.

        :param path: SHV path to the property node.
        :param value: The value we compare against. It is ignored for if we have this
            path in cache and is ``None``. Otherwise it is used to actually detect the
            change (if returned value is not equal to this one).
        :param timeout: How long we should wait for change. Pass ``None`` to wait
            infinitely.
        :param get_period: How often the pooling should be performed.
        """
        tasks: set[asyncio.Task] = {
            asyncio.create_task(
                self._prop_change_wait(
                    path,
                    self.get(path, None) if value is None else value,
                    get_period,
                )
            )
        }
        if self.is_subscribed(path):
            tasks.add(asyncio.create_task(self.wait_for_change(path)))
        done, pending = await asyncio.wait(
            tasks, timeout=timeout, return_when=asyncio.FIRST_COMPLETED
        )
        for task in pending:
            task.cancel()
        for task in pending:
            try:
                await task
            except asyncio.CancelledError:
                pass
        if not done:
            raise TimeoutError
        return done.pop().result()  # type: ignore

    async def _prop_change_wait(
        self, path: str, value: SHVType, period: float
    ) -> SHVType:
        while True:
            v = await self.prop_get(path)
            if v != value:
                return v
            await asyncio.sleep(period)

    def on_change(
        self,
        path: str,
        callback: collections.abc.Callable[["ValueClient", str, SHVType], None] | None,
    ) -> None:
        """Register callback handler called when value change is reported.

        The handler is called right before value is updated and thus it is possible to
        access the old and new value is provided as an argument. Note that it is up to
        the device if this signal is sent really only on value change or if it is sent
        more often.

        The default implementation of handler lookup (:meth:`_get_handler`) is that the
        most matching callback path is selected and callback for it called. This way you
        can get all notifications delivered to a single handler if you use empty string
        as the path.

        :param path: SHV path to the node (includes its children) change is expected on.
        :param callback: Function called when change notification is received. You can
            pass ``None`` to remove any existing callback. Note that there can be only
            one callback registered for a single path and thus a different callback
            replaces the previous one.
        """
        if callback is None:
            self._handlers.pop(path)
        else:
            self._handlers[path] = callback

    async def wait_for_change(self, path: str) -> SHVType:
        """Provide a way to await for the future change signal.

        Compared to the :meth:`on_change` the path has to match exactly.

        :param path: SHV path to the node the change signal is expected on.
        :return: future value.
        """
        if path not in self._futures:
            self._futures[path] = []
        future: asyncio.Future[SHVType] = asyncio.get_running_loop().create_future()
        self._futures[path].append(future)
        return await future

    async def unsubscribe(self, sub: str, clean_cache: bool = True) -> bool:
        """Perform unsubscribe for signals on given path.

        :param sub: SHV RPC subscription to be removed.
        :param wipe_cache: If no longer subscribed paths should be removed from cache or
            not. The default is to remove them but you can also do multiple unsibscribes
            and then call :meth:`clean_cache` for all of the at once.
        :return: ``True`` in case such subscribe was located and ``False`` otherwise.
        """
        res = await super().unsubscribe(sub)
        if res and clean_cache:
            self.clean_cache()
        return res

    def is_subscribed(
        self, path: str, signal: str = "chng", source: str = "get"
    ) -> bool:
        """Check if we are subscribed for given SHV path.

        Subscribed paths are cached and thus this also checks if this path would be
        cached.

        This is only local check. This won't reach the server to verify that all
        subscriptions are still valid (not removed on the server).

        :param path: SHV path
        :param signal: Signal name
        :param source: Method name signal is associated with.
        :return: ``True`` if subscribed for that path and ``False`` otherwise.
        """
        return any(rpcri_match(sub, path, source, signal) for sub in self._subscribes)

    def clean_cache(self) -> None:
        """Remove no longer subscribed paths from cache.

        There is commonly no need to call this method unless you call
        :meth:`unsubscribe` with ``wipe_cache=False``.
        """
        self._cache = {k: v for k, v in self._cache.items() if self.is_subscribed(k)}

    async def log_snapshot(self, path: str) -> None:
        """Get snapshot of the logs.

        Use this to receive old values.

        :param path: SHV path.
        """
        param: SHVMapType = {
            "recordCountLimit": 10000,
            "withPathsDict": True,
            "withSnapshot": True,
            "withTypeInfo": False,
            "since": datetime.datetime.now(),
        }
        result = await self.call(path, "getLog", param)
        if result:
            paths_dict = shvmeta(result).get("pathsDict", None)
            if isinstance(paths_dict, collections.abc.Sequence):
                for list_item in paths_dict:
                    if not isinstance(list_item, collections.abc.Sequence):
                        continue
                    idx = list_item[1]
                    if not isinstance(idx, int):
                        continue
                    value = list_item[2]
                    spath = paths_dict[idx]
                    if not isinstance(spath, str):
                        continue
                    await self._value_update(spath, value)

    async def get_snapshot(self, *paths: str, update: bool = False) -> None:
        """Get snapshot of data on subscribed paths using get methods.

        This provides a way for you to initialize cache without logs. It
        iterates over SHV tree and calls any get method it encounters.

        :param paths: Paths to be snapshoted. If none is provide the
          subscriptions are used instead.
        :param update: If already cached values should be updated or just
          skipped.
        """
        pths: list[str] = list(paths) if paths else [""]
        # TODO we can skip paths that are outside of our subscriptions
        while pths:
            pth = pths.pop()
            try:
                pths.extend(
                    f"{pth}{'/' if pth else ''}{name}" for name in await self.ls(pth)
                )
            except (RpcMethodNotFoundError, RpcMethodCallExceptionError):
                pass  # ls might not be present which is not an issue
            if not self.is_subscribed(pth) or (not update and pth in self._cache):
                continue
            if await self.dir_exists(pth, "get"):
                self._cache[pth] = (time.time(), await self.prop_get(pth))



"""Implementation of RpcMessage."""

from __future__ import annotations

import collections.abc
import enum
import typing

from .chainpack import ChainPackWriter
from .cpon import CponWriter
from .rpcerrors import RpcError, RpcErrorCode
from .rpcmethod import RpcMethodAccess
from .value import SHVIMap, SHVIMapType, SHVType, is_shvimap, shvmeta_eq


class RpcMessage:
    """Single SHV RPC message representation."""

    last_request_id: typing.ClassVar[int] = 0
    """Counter of request IDs to ensure that every request has unique ID."""

    @classmethod
    def next_request_id(cls) -> int:
        """Provide unique request identifier.

        The identifier won't repeat for this application as it is just simple
        counter that should never wrap.
        """
        cls.last_request_id += 1
        return cls.last_request_id

    def __init__(self, rpc_val: SHVIMapType | None = None) -> None:
        if rpc_val is None:
            rpc_val = SHVIMap()
        if not isinstance(rpc_val, SHVIMap):
            rpc_val = SHVIMap(rpc_val)
        self.value: SHVIMap = rpc_val

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcMessage) and shvmeta_eq(self.value, other.value)

    def __repr__(self) -> str:
        return f"<RpcMessage {self.value.meta!r}: {self.value!r}>"

    class Tag(enum.IntEnum):
        """Tags in Meta for RPC message."""

        REQUEST_ID = 8
        PATH = 9
        METHOD = 10
        SIGNAL = 10
        CALLER_IDS = 11
        RESP_CALLER_IDS = 13
        ACCESS = 14
        USER_ID = 16
        ACCESS_LEVEL = 17
        PART = 18
        SOURCE = 19

    class Key(enum.IntEnum):
        """Keys in the toplevel IMap of the RPC message."""

        PARAMS = 1
        RESULT = 2
        ERROR = 3

    class ErrorKey(enum.IntEnum):
        """Keys in the error IMap."""

        CODE = 1
        MESSAGE = 2

    def is_valid(self) -> bool:
        """Check if message is valid RPC message."""
        return (
            isinstance(self.value, SHVIMap)
            and bool(set(self.value) ^ set(self.Key))
            and (self.is_request or self.is_response or self.is_signal)
            and isinstance(self._path, str)
            and isinstance(self._signal_name, str)
            and isinstance(self._source, str)
        )

    @property
    def is_request(self) -> bool:
        """Check if message is request."""
        return bool(self.has_request_id and self.has_method)

    @property
    def is_response(self) -> bool:
        """Check if message is a response."""
        return bool(self.has_request_id and not self.has_method)

    @property
    def is_error(self) -> bool:
        """Check if message is an error response."""
        return bool(
            self.is_response
            and self.error is not None
            and self.rpc_error.error_code != RpcErrorCode.NO_ERROR
        )

    @property
    def is_signal(self) -> bool:
        """Check if message is a signal."""
        return not self.has_request_id

    def make_response(
        self, result: SHVType = None, error: RpcError | SHVType = None
    ) -> RpcMessage:
        """Create new message that is response to this one.

        :param result: The result value to be set in the response.
        :param error: The error to be set to the response message. You should
          use either *result* or *error* not both.
        :return: The new message that is response to this one.
        """
        if not self.is_request:
            raise ValueError("Response can be created from request only.")
        resp = RpcMessage()
        resp.request_id = self.request_id
        resp.caller_ids = self.caller_ids
        resp.result = result
        if isinstance(error, RpcError):
            resp.rpc_error = error
        else:
            resp.error = error
        return resp

    @property
    def has_request_id(self) -> bool:
        """Check if valid request ID was provided in this message."""
        return self.Tag.REQUEST_ID in self.value.meta and isinstance(
            self.value.meta[self.Tag.REQUEST_ID], int
        )

    @property
    def request_id(self) -> int:
        """Request identificator of this message."""
        res = self.value.meta[self.Tag.REQUEST_ID]
        if not isinstance(res, int):
            raise ValueError(f"Invalid request ID type: {type(res)}")
        return res

    @request_id.setter
    def request_id(self, rqid: int | None) -> None:
        """Set given request identicator to this message."""
        if rqid is None:
            self.value.meta.pop(self.Tag.REQUEST_ID, None)
        else:
            self.value.meta[self.Tag.REQUEST_ID] = rqid

    def new_request_id(self) -> int:
        """Set new request ID.

        :return: The new request ID.
        """
        self.request_id = self.next_request_id()
        return self.request_id

    @property
    def _path(self) -> SHVType:
        return self.value.meta.get(self.Tag.PATH, "")

    @property
    def path(self) -> str:
        """SHV path specified for this message or empty string."""
        res = self._path
        if not isinstance(res, str):
            raise ValueError(f"Invalid path type: {type(res)}")
        return res

    @path.setter
    def path(self, path: str) -> None:
        """Set given path as SHV path for this message."""
        if path:
            self.value.meta[self.Tag.PATH] = path
        else:
            self.value.meta.pop(self.Tag.PATH, None)

    @property
    def has_method(self) -> bool:
        """Check if valid method name was provided in this message."""
        return self.Tag.METHOD in self.value.meta and isinstance(
            self.value.meta[self.Tag.METHOD], str
        )

    @property
    def method(self) -> str:
        """SHV method name for this message."""
        res = self.value.meta[self.Tag.METHOD]
        if not isinstance(res, str):
            raise ValueError(f"Invalid method type: {type(res)}")
        return res

    @method.setter
    def method(self, method: str) -> None:
        """Set SHV method name for this message."""
        if method:
            self.value.meta[self.Tag.METHOD] = method
        else:
            self.value.meta.pop(self.Tag.METHOD, None)

    @property
    def _signal_name(self) -> SHVType:
        return self.value.meta.get(self.Tag.METHOD, "chng")

    @property
    def signal_name(self) -> str:
        """SHV signal name for this message."""
        res = self._signal_name
        if not isinstance(res, str):
            raise ValueError(f"Invalid method type: {type(res)}")
        return res

    @signal_name.setter
    def signal_name(self, signal: str) -> None:
        """Set SHV signal name for this message."""
        # Note: we always set it because old implementations were dropping
        # messages without method name.
        self.value.meta[self.Tag.SIGNAL] = signal

    @property
    def _source(self) -> SHVType:
        return self.value.meta.get(self.Tag.SOURCE, "get")

    @property
    def source(self) -> str:
        """SHV signal source method name for this message."""
        res = self._source
        if not isinstance(res, str):
            raise ValueError(f"Invalid method type: {type(res)}")
        return res

    @source.setter
    def source(self, source: str) -> None:
        """Set SHV signal source method name for this message."""
        if source and source != "get":
            self.value.meta[self.Tag.SOURCE] = source
        else:
            self.value.meta.pop(self.Tag.SOURCE, None)

    @property
    def caller_ids(self) -> collections.abc.Sequence[int]:
        """Caller idenfieiers associated with this message."""
        res = self.value.meta.get(self.Tag.CALLER_IDS, None)
        if isinstance(res, int):
            return [res]
        if not isinstance(res, list):
            return []
        return list(filter(lambda v: isinstance(v, int), res))

    @caller_ids.setter
    def caller_ids(self, cids: collections.abc.Sequence[int]) -> None:
        """Set caller idenfieiers associated with this message."""
        if not cids:
            self.value.meta.pop(self.Tag.CALLER_IDS, None)
        else:
            some: SHVType = cids
            if len(cids) == 1:
                some = cids[0]
            self.value.meta[self.Tag.CALLER_IDS] = some

    @property
    def access(self) -> collections.abc.Sequence[str]:
        """Granted access sequence."""
        res = self.value.meta.get(self.Tag.ACCESS, "")
        if isinstance(res, str):
            if res:
                return res.split(",")
        return []

    @access.setter
    def access(self, access: collections.abc.Sequence[str]) -> None:
        """Set granted access sequence."""
        if access:
            self.value.meta[self.Tag.ACCESS] = access
        else:
            self.value.meta.pop(self.Tag.ACCESS, None)

    @property
    def rpc_access(self) -> RpcMethodAccess | None:
        """Access level as :class:`shv.RpcMethodAccess`."""
        m = RpcMethodAccess.strmap()
        if level := self.value.meta.get(self.Tag.ACCESS_LEVEL):
            if isinstance(level, int):
                return RpcMethodAccess(level)
        for access in self.access:
            if access in m:
                return m[access]
        return None

    @rpc_access.setter
    def rpc_access(self, access: RpcMethodAccess | None) -> None:
        """Set access level with :class:`shv.RpcMethodAccess`."""
        if access is not None:
            self.value.meta[self.Tag.ACCESS] = RpcMethodAccess.tostr(access)
            self.value.meta[self.Tag.ACCESS_LEVEL] = access.value
        else:
            self.value.meta.pop(self.Tag.ACCESS, None)

    @property
    def user_id(self) -> str | None:
        """User's ID caried by message."""
        res = self.value.meta.get(self.Tag.USER_ID, None)
        if isinstance(res, dict):  # Note: backward compatibility
            res = f"{res.get('brokerId')}:{res.get('shvUser')}"
        if res is not None and not isinstance(res, str):
            res = str(res)
        return res

    @user_id.setter
    def user_id(self, value: str | None) -> None:
        """Set User's ID."""
        if value is not None:
            self.value.meta[self.Tag.USER_ID] = value
        else:
            self.value.meta.pop(self.Tag.USER_ID, None)

    @property
    def param(self) -> SHVType:
        """SHV parameters for the method call."""
        return self.value.get(self.Key.PARAMS, None) if is_shvimap(self.value) else None

    @param.setter
    def param(self, param: SHVType) -> None:
        """Set SHV parameters for this method call."""
        if param is None:
            self.value.pop(self.Key.PARAMS, None)
        else:
            self.value[self.Key.PARAMS] = param

    @property
    def result(self) -> SHVType:
        """SHV method call result."""
        return self.value.get(self.Key.RESULT, None) if is_shvimap(self.value) else None

    @result.setter
    def result(self, result: SHVType) -> None:
        """Set SHV method call result."""
        if result is None:
            self.value.pop(self.Key.RESULT, None)
        else:
            self.value[self.Key.RESULT] = result

    @property
    def error(self) -> SHVType:
        """SHV method call error."""
        return self.value.get(self.Key.ERROR, None) if is_shvimap(self.value) else None

    @error.setter
    def error(self, error: SHVType) -> None:
        """Set SHV method call error."""
        if error is None:
            self.value.pop(self.Key.ERROR, None)
        else:
            self.value[self.Key.ERROR] = error

    @property
    def rpc_error(self) -> RpcError:
        """SHV method call error in standard SHV format :class:`RpcError`."""
        res = self.error
        if is_shvimap(res):
            rcode = res.get(self.ErrorKey.CODE)
            code: RpcErrorCode = RpcErrorCode.UNKNOWN
            if isinstance(rcode, int):
                try:
                    code = RpcErrorCode(rcode)
                except ValueError:
                    pass
            rmsg = res.get(self.ErrorKey.MESSAGE)
            msg = rmsg if isinstance(rmsg, str) else ""
            return RpcError(msg, code)
        return RpcError(res if isinstance(res, str) else "", RpcErrorCode.UNKNOWN)

    @rpc_error.setter
    def rpc_error(self, error: RpcError) -> None:
        """Set SHV method call error in standard SHV format."""
        if error.error_code == RpcErrorCode.NO_ERROR:
            self.error = None
        else:
            err: SHVIMapType = {
                self.ErrorKey.CODE: error.error_code,
                **(
                    {}
                    if error.message is None
                    else {self.ErrorKey.MESSAGE: error.message}
                ),
            }
            self.error = err

    def to_string(self) -> str:
        """Convert message to CPON and return it as string."""
        return self.to_cpon().decode("utf-8")

    def to_cpon(self) -> bytes:
        """Convert message to Cpon."""
        return CponWriter.pack(self.value)

    def to_chainpack(self) -> bytes:
        """Convert message to Chainpack."""
        return ChainPackWriter.pack(self.value)

    @classmethod
    def request(
        cls,
        path: str,
        method: str,
        param: SHVType = None,
        rid: int | None = None,
        user_id: str | None = None,
    ) -> RpcMessage:
        """Create request message.

        :param path: SHV path for signal.
        :param method: method name for signal.
        :param param: Parameters passed to the method.
        :param rid: Request identifier for this message. It is automatically assigned if
          ``None`` is passed.
        :param user_id: User's ID to be caried with request message.
        """
        res = cls()
        res.request_id = rid or cls.next_request_id()
        res.method = method
        res.path = path
        res.param = param
        res.user_id = user_id
        return res

    @classmethod
    def signal(
        cls,
        path: str,
        name: str = "chng",
        source: str = "get",
        value: SHVType = None,
        access: RpcMethodAccess = RpcMethodAccess.READ,
    ) -> RpcMessage:
        """Create signal message.

        :param path: SHV path for signal.
        :param name: Name of the signal.
        :param source: Name of the method this signal is associated with.
        :param value: Value to be sent in the message.
        :param access: Minimal access level needed to get this signal.
        """
        res = cls()
        res.signal_name = name
        res.source = source
        res.path = path
        res.param = value
        res.rpc_access = access
        return res



"""The smarter loader of the version.

It identifies if this package is installed and returns version from egg info or
it looks for pyproject.toml file in upper directory.
"""

import importlib.metadata
import pathlib
import tomllib


def _get_version() -> str:
    for dist_name in importlib.metadata.packages_distributions().get(__package__, []):
        dist = importlib.metadata.distribution(dist_name)
        if str(dist.locate_file(f"{__package__}/__version__.py")) == __file__:
            return dist.version
    pyproject = pathlib.Path(__file__).parent.parent / "pyproject.toml"
    if pyproject.exists():
        with pyproject.open("rb") as f:
            res = tomllib.load(f)["project"]["version"]
            if isinstance(res, str):
                return res
    return "unknown"


VERSION = _get_version()



"""The implementation of conversion."""

import enum
import io
import pathlib
import typing

from ..chainpack import ChainPackReader, ChainPackWriter
from ..commonpack import CommonReader, CommonWriter
from ..cpon import CponReader, CponWriter


class CPFormat(enum.Enum):
    """The specified for the format we should convert to."""

    CPON = enum.auto()
    CHAINPACK = enum.auto()


def convert(
    inp: str | bytes | pathlib.Path | typing.IO,
    inf: CPFormat,
    outp: None | pathlib.Path | typing.IO,
    outf: CPFormat,
    cpon_options: CponWriter.Options | None = None,
) -> bytes | None:
    """Convert ChainPack to Cpon or vice versa.

    This also allows conversion from the same type to the same one. This can be
    used to validate format for example.

    :param inp: Input data in ChainPack or Cpon or path to the file with them.
    :param inf: Input data format.
    :param outp: Ouput where `None` results in data returned from the function
        and file or IO into the write to those.
    :param outf: Output data format.
    :param options: Cpon writer options. This is used only when:
        `outf == CPFormat.CPON`.
    :return: Bytes in case `odata` was `None` otherwise `None`.
    """
    if isinstance(inp, str):
        inp = inp.encode("utf-8")

    istream: typing.IO
    if isinstance(inp, bytes):
        istream = io.BytesIO(inp)
    elif isinstance(inp, pathlib.Path):
        istream = inp.open("rb")
    else:
        istream = inp

    ostream: typing.IO
    if outp is None:
        ostream = io.BytesIO()
    elif isinstance(outp, pathlib.Path):
        ostream = outp.open("wb")
    else:
        ostream = outp

    rd: CommonReader = (
        CponReader(istream) if inf is CPFormat.CPON else ChainPackReader(istream)
    )
    wr: CommonWriter = (
        CponWriter(ostream, cpon_options)
        if outf is CPFormat.CPON
        else ChainPackWriter(ostream)
    )

    wr.write(rd.read())

    if isinstance(ostream, io.BytesIO):
        return ostream.getvalue()
    if isinstance(outp, pathlib.Path):
        ostream.close()
    return None



"""RPC File access and implementation utility."""

from __future__ import annotations

import binascii
import collections
import dataclasses
import datetime
import enum
import hashlib
import io
import pathlib
import types
import typing

from .rpcerror import RpcInvalidParamsError
from .rpcmethod import RpcMethodAccess, RpcMethodDesc, RpcMethodFlags
from .rpcparams import shvargt, shvgett, shvt
from .simplebase import SimpleBase
from .value import SHVType, is_shvimap


@dataclasses.dataclass
class RpcFileStat:
    """The stat information for the RPC File."""

    size: int
    """Size of the file in bytes."""
    page_size: int
    """Page size (ideal size and thus alignment for efficient access)."""
    access_time: datetime.datetime | None = None
    """Optional information about the latest data access."""
    mod_time: datetime.datetime | None = None
    """Optional information about the latest data modification."""

    class Key(enum.IntEnum):
        """Key in the stat IMap."""

        TYPE = 0
        SIZE = 1
        PAGE_SIZE = 2
        ACCESS_TIME = 3
        MOD_TIME = 4

    def to_shv(self) -> SHVType:
        """Convert to SHV RPC representation."""
        res: dict[int, SHVType] = {
            self.Key.TYPE: 0,
            self.Key.SIZE: self.size,
            self.Key.PAGE_SIZE: self.page_size,
        }
        if self.access_time is not None:
            res[self.Key.ACCESS_TIME] = self.access_time
        if self.mod_time is not None:
            res[self.Key.MOD_TIME] = self.mod_time
        return res

    @classmethod
    def from_shv(cls, value: SHVType) -> RpcFileStat:
        """Create from SHV RPC representation."""
        if not is_shvimap(value):
            raise ValueError("Expected Map.")
        if value.get(cls.Key.TYPE, 0) != 0:
            raise ValueError("Unsupported type")
        return cls(
            size=shvgett(value, cls.Key.SIZE, int, 0),
            page_size=shvgett(value, cls.Key.PAGE_SIZE, int, 128),
            access_time=shvgett(value, cls.Key.ACCESS_TIME, datetime.datetime, None),
            mod_time=shvgett(value, cls.Key.MOD_TIME, datetime.datetime, None),
        )

    @classmethod
    def for_path(cls, path: pathlib.Path | str) -> RpcFileStat:
        """Create stat information for the existing file.

        :param path: Path to the file for which stat should be provided.
        :return: The RPC stat object.
        :raises FileNotFoundError: If there is no such file or if path exists
          but doesn't lead to the file.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)
        if not path.is_file():
            raise FileNotFoundError("Not a regular file")
        stat = path.stat()
        return cls(
            stat.st_size,
            stat.st_blksize,
            datetime.datetime.fromtimestamp(stat.st_atime),
            datetime.datetime.fromtimestamp(stat.st_mtime),
        )


class RpcFile:
    """RPC file accessed over :class:`SimpleBase`.

    :param client:
    :param path:
    :param buffer_read:
    :param buffer_write:
    """

    def __init__(
        self,
        client: SimpleBase,
        path: str,
        buffered: bool = True,
        append: bool = False,
    ) -> None:
        self.client: SimpleBase = client
        """The client used to access the RPC file."""
        self.offset: int = 0
        """Access offset of the file.

        This gets modified by reading and writing from the file. You can also
        set it to perform seek.
        """
        self._path = path
        self._append = False
        self._read_page: int = 0
        self._read_buffer: bytearray | None = bytearray() if buffered else None
        self._write_offset: int = 0
        self._write_buffer: bytearray | None = bytearray() if buffered else None
        self.__page_size: int | None = None

    @property
    def path(self) -> str:
        """SHV path to the file this object is used to access."""
        return self._path

    async def __aenter__(self) -> RpcFile:
        self.offset = 0
        return self

    async def __aexit__(
        self,
        exc_type: type[Exception],
        exc_value: Exception,
        traceback: types.TracebackType,
    ) -> None:
        await self.flush()
        self._read_buffer = bytearray()

    async def __aiter__(self) -> RpcFile:
        return self

    async def __anext__(self) -> bytes:
        res = await self.readline()
        if not res:
            raise StopAsyncIteration
        return res

    async def readable(self) -> bool:
        """Check if file is readable."""
        return await self.client.dir_exists(self._path, "read")

    async def writable(self) -> bool:
        """Check if file is writable with selected write method."""
        return await self.client.dir_exists(
            self._path, "append" if self._append else "write"
        )

    async def resizable(self) -> bool:
        """Check if file is resizable."""
        return await self.client.dir_exists(self._path, "truncate")

    async def stat(self) -> RpcFileStat:
        """Get the file stat info."""
        return RpcFileStat.from_shv(await self.client.call(self._path, "stat"))

    async def _page_size(self) -> int:
        """Get the ideal page size."""
        if self.__page_size is None:
            self.__page_size = (await self.stat()).page_size
        return self.__page_size

    async def read(self, size: int = -1) -> bytes:
        """Read at most given number of bytes.

        :param size: Maximum number of bytes to read. The value 0 or less is the
          no size limit.
        """
        page_size = await self._page_size()
        res = bytearray()
        if self._read_buffer is None:
            while True:
                cnt = size if 0 < size < page_size else page_size
                data = await self.client.call(self._path, "read", [self.offset, cnt])
                if not isinstance(data, bytes) or not data:
                    break
                res += data
                self.offset += len(data)
                if size > 0:
                    size -= len(data)
                    if size == 0:
                        break
        else:
            pass
        return bytes(res)

    async def readline(self, size: int = -1) -> bytes:
        """Read and return one line from the file from the current offset."""
        if self._read_buffer is None:
            raise NotImplementedError(
                "Reading line by line is only supported with buffering"
            )
        res = bytearray()
        while size != 0:
            b = self.read(1)
            if not b:
                break

        return bytes(res)

    async def crc32(self, offset: int | None = 0, size: int | None = None) -> int:
        """Calculate CRC32 checksum of the data.

        :param offset: offset from the file start or current file offset if
          ``None`` is passed.
        :param size: number of bytes since offset used to calculate checksum or
          all of them up to the end of the file in case ``None`` is passed.
        :return: Calculated CRC32 value.
        """
        if offset == 0 and size is None:
            param = None
        else:
            param = [self.offset if offset is None else offset, size]
        res = await self.client.call(self._path, "crc", param)
        return res if isinstance(res, int) else 0

    async def sha1(self, offset: int | None = 0, size: int | None = None) -> bytes:
        """Calculate SHA1 checksum of the data.

        :param offset: offset from the file start or current file offset if
          ``None`` is passed.
        :param size: number of bytes since offset used to calculate checksum or
          all of them up to the end of the file in case ``None`` is passed.
        :return: Calculated SHA1 value.
        """
        if offset == 0 and size is None:
            param = None
        else:
            param = [self.offset if offset is None else offset, size]
        res = await self.client.call(self._path, "sha1", param)
        return res if isinstance(res, bytes) else bytes(20)

    async def _page_write(self, offset: int, data: bytes | bytearray) -> None:
        page_size = await self._page_size()
        while len(data) > page_size:
            param = [self.offset, data[:page_size]]
            await self.client.call(self._path, "write", param)
            data = data[page_size:]

    async def write(self, data: bytes | bytearray) -> None:
        """Write data to the file on the current offset."""
        if self._write_buffer is None:
            if self._append:
                await self.client.call(self._path, "append", data)
            else:
                await self._page_write(self.offset, data)
            return

        if self._append:
            self._write_buffer += data
            return

        page_size = await self._page_size()
        if self._write_offset + len(self._write_buffer) != self.offset:
            await self._page_write(self._write_offset, self._write_buffer)
            self._write_buffer = bytearray()
        if not self._write_buffer:
            self._write_offset = self.offset
        self._write_buffer += data
        if len(self._write_buffer) > page_size:
            # Write what we have so far
            pass

    async def flush(self) -> None:
        """Flush data buffered for write by previous :meth:`write` calls."""
        if self._write_buffer is None:
            return
        if self._append:
            await self.client.call(self._path, "append", self._write_buffer)
        else:
            await self._page_write(self._write_offset, self._write_buffer)
        self._write_buffer = bytearray()

    async def truncate(self, size: int | None = None) -> None:
        """Truncate file size.

        :param size: Truncate to this specific size or to the current offset in
        case of ``None``.
        """
        await self.client.call(
            self._path, "truncate", size if size is not None else self.offset
        )


class RpcTextFile:
    """RPC text file accessed over :class:`SimpleBase`.

    This is based on :class:`RpcFile` with addition that it interprets received
    data as text instead of bytes.
    """

    def __init__(self, file: RpcFile) -> None:
        self.file = file
        raise NotImplementedError

    # TODO


@dataclasses.dataclass
class FileProvider:
    """Implementation helper that provides files over SHV RPC."""

    access_read: RpcMethodAccess = RpcMethodAccess.READ
    """Access level for ``read`` method."""
    access_write: RpcMethodAccess | None = RpcMethodAccess.WRITE
    """Access level for ``write`` method."""
    access_truncate: RpcMethodAccess | None = RpcMethodAccess.WRITE
    """Access level for ``truncate`` method."""
    access_append: RpcMethodAccess | None = RpcMethodAccess.WRITE
    """Access level for ``append`` method."""

    def dir(self) -> collections.abc.Iterator[RpcMethodDesc]:
        """Provide method descriptions for this file access."""
        yield RpcMethodDesc.getter("stat", "Null", "ofstat", self.access_read)
        yield RpcMethodDesc.getter("size", "Null", "Int", self.access_read)
        yield RpcMethodDesc(
            "crc", param="ifcrc", result="UInt", access=self.access_read
        )
        yield RpcMethodDesc(
            "sha1", param="ifsha1", result="Bytes", access=self.access_read
        )
        yield RpcMethodDesc(
            "read",
            RpcMethodFlags.LARGE_RESULT_HINT,
            "ifread",
            "Bytes",
            self.access_read,
        )
        if self.access_write is not None:
            yield RpcMethodDesc(
                "write", param="ifwrite", result="Null", access=self.access_write
            )
        if self.access_truncate is not None:
            yield RpcMethodDesc(
                "truncate", param="Int", result="Null", access=self.access_truncate
            )
        if self.access_append is not None:
            yield RpcMethodDesc(
                "append", param="Bytes", result="Null", access=self.access_append
            )

    def method(
        self,
        path: pathlib.Path,
        name: str,
        param: SHVType,
        access_level: RpcMethodAccess,
    ) -> SHVType:
        """File access methods implementation.

        :param path: Path to the file that should be accessed. This is not SHV
          path but rather a real local file path.
        :param name: Name of the method that is called.
        :param param: Parameter provided to the method call.
        :param access_level: The access level for the method call.
        :return: The value that should be provided as result of the method call.
        :raises FileNotFoundError: If file pointed to by ``path`` doesn't
          exists.
        :raises RpcInvalidParamsError: If parameter is invalid.
        :raises NotImplementedError: If requested method is not implemented or
          too low access level was provided.
        """
        match name:
            case "stat" if access_level >= self.access_read:
                return RpcFileStat.for_path(path).to_shv()
            case "size" if access_level >= self.access_read:
                return path.stat().st_size
            case "crc" if access_level >= self.access_read:
                offset = shvargt(param, 0, int, 0)
                size = shvargt(param, 1, int, None)
                crc = binascii.crc32(b"")
                with path.open("rb") as file:
                    file.seek(offset)
                    while size is None or size > 0:
                        data = file.read(
                            io.DEFAULT_BUFFER_SIZE
                            if size is None
                            else min(size, io.DEFAULT_BUFFER_SIZE)
                        )
                        if not data:
                            break
                        crc = binascii.crc32(data, crc)
                        if size is not None:
                            size -= len(data)
                return crc
            case "sha1" if access_level >= self.access_read:
                offset = shvargt(param, 0, int, 0)
                size = shvargt(param, 1, int, None)
                hash = hashlib.sha1()  # noqa PLR6301
                with path.open("rb") as file:
                    file.seek(offset)
                    while size is None or size > 0:
                        data = file.read(
                            io.DEFAULT_BUFFER_SIZE
                            if size is None
                            else min(size, io.DEFAULT_BUFFER_SIZE)
                        )
                        if not data:
                            break
                        hash.update(data)
                        if size is not None:
                            size -= len(data)
                return hash.digest()
            case "read" if access_level >= self.access_read:
                offset = shvargt(param, 0, int)
                size = shvargt(param, 1, int)
                with path.open("rb") as file:
                    file.seek(offset)
                    return file.read(size)
            case "write" if self.access_write is not None and access_level >= self.access_write:
                offset = shvargt(param, 0, int)
                data = shvargt(param, 1, bytes)
                if (
                    self.access_truncate is None
                    and (offset + len(data)) > path.stat().st_size
                ):
                    raise RpcInvalidParamsError(
                        "Write beyond the file boundary is not possible"
                    )
                with path.open("wb") as file:
                    file.seek(offset)
                    file.write(data)
                return None
            case "truncate" if self.access_truncate is not None and access_level >= self.access_truncate:
                size = shvt(param, int)
                with path.open("wb") as file:
                    file.truncate(size)
                return None
            case "append" if self.access_append is not None and access_level >= self.access_append:
                data = shvt(param, bytes)
                with path.open("ab") as file:
                    file.write(data)
                return None
        raise NotImplementedError


FileProviderRO: typing.Final = FileProvider(
    access_read=RpcMethodAccess.READ,
    access_write=None,
    access_truncate=None,
    access_append=None,
)
"""RPC File provider for read-only file access."""
FileProviderRW: typing.Final = FileProvider(
    access_read=RpcMethodAccess.READ,
    access_write=RpcMethodAccess.WRITE,
    access_truncate=RpcMethodAccess.WRITE,
    access_append=RpcMethodAccess.WRITE,
)
"""RPC File provider for read-write access to the files."""
FileProviderFixedSize: typing.Final = FileProvider(
    access_read=RpcMethodAccess.READ,
    access_write=RpcMethodAccess.WRITE,
    access_truncate=None,
    access_append=None,
)
"""RPC File provider for read-write access to the files of fixed size."""
FileProviderAppend: typing.Final = FileProvider(
    access_read=RpcMethodAccess.READ,
    access_write=None,
    access_truncate=None,
    access_append=RpcMethodAccess.WRITE,
)
"""RPC File provider that aprovides write access but only through appends."""



"""We are using :func:`binascii.crc32` but check if it is really CRC-32."""

import binascii

import pytest


@pytest.mark.parametrize(
    "data,crc",
    (
        (b"", 0x0),
        (b"1", 0x83DCEFB7),
        (b"123456789", 0xCBF43926),
    ),
)
def test(data, crc):
    """Test that :func:`binascii.crc32` provides implementation we need."""
    assert binascii.crc32(data) == crc



"""Implementation of Rpc RPC specific errors."""

from __future__ import annotations

import dataclasses
import enum


class RpcErrorCode(enum.IntEnum):
    """Number representing an SHV RPC error code."""

    NO_ERROR = 0
    INVALID_REQUEST = 1
    METHOD_NOT_FOUND = 2
    INVALID_PARAM = 3
    INTERNAL_ERR = 4
    PARSE_ERR = 5
    METHOD_CALL_TIMEOUT = 6
    METHOD_CALL_CANCELLED = 7
    METHOD_CALL_EXCEPTION = 8
    UNKNOWN = 9
    LOGIN_REQUIRED = 10
    USER_ID_REQUIRED = 11
    NOT_IMPLEMENTED = 12
    USER_CODE = 32


class RpcError(RuntimeError):
    """Top level for Rpc RPC errors.

    This tries to be a bit smart when objects are created. There is class
    property `shv_error_map` that is used to lookup a more appropriate error
    object. If you are adding any of your own error codes you can add them to
    this map to be looked up by their code.

    :param msg: Message describing the error circumstances.
    :param code: Error code.
    """

    shv_error_code: RpcErrorCode = RpcErrorCode.UNKNOWN
    shv_error_map: dict[int, type[RpcError]] = dataclasses.field(default_factory=dict)

    def __new__(
        cls, msg: str | None = None, code: RpcErrorCode | None = None
    ) -> RpcError:
        """Create an appropriate exception based on the code."""
        ncls = cls.shv_error_map.get(cls.shv_error_code if code is None else code, cls)
        return super(RpcError, cls).__new__(ncls)  # noqa UP008

    def __init__(
        self, msg: str | None = None, code: RpcErrorCode | None = None
    ) -> None:
        super().__init__(msg, self.shv_error_code if code is None else code)

    @property
    def message(self) -> str | None:
        """Provides access to the SHV RPC message."""
        if not isinstance(self.args[0], str) and self.args[0] is not None:
            raise ValueError(f"Must be string or None but is: {type(self.args[0])}")
        return self.args[0]

    @property
    def error_code(self) -> RpcErrorCode:
        """Provides access to the :class:`RpcErrorCode`."""
        assert isinstance(self.args[1], RpcErrorCode)
        return self.args[1]


class RpcInvalidRequestError(RpcError):
    """The sent data are not valid request object."""

    shv_error_code = RpcErrorCode.INVALID_REQUEST


class RpcMethodNotFoundError(RpcError):
    """The method does not exist or is not available."""

    shv_error_code = RpcErrorCode.METHOD_NOT_FOUND


class RpcInvalidParamError(RpcError):
    """Invalid method parameter were provided."""

    shv_error_code = RpcErrorCode.INVALID_PARAM


class RpcInternalError(RpcError):
    """Internal JSON-RPC error."""

    shv_error_code = RpcErrorCode.INTERNAL_ERR


class RpcParseError(RpcError):
    """Invalid data were recived and were not possible to parse them."""

    shv_error_code = RpcErrorCode.PARSE_ERR


class RpcMethodCallTimeoutError(RpcError):
    """Method call timed out without providing result."""

    shv_error_code = RpcErrorCode.METHOD_CALL_TIMEOUT


class RpcMethodCallCancelledError(RpcError):
    """Method call was cancelled."""

    shv_error_code = RpcErrorCode.METHOD_CALL_CANCELLED


class RpcMethodCallExceptionError(RpcError):
    """Method call resulted in exception and not result."""

    shv_error_code = RpcErrorCode.METHOD_CALL_EXCEPTION


class RpcLoginRequiredError(RpcError):
    """Login sequence must be performed before anything else."""

    shv_error_code = RpcErrorCode.LOGIN_REQUIRED


class RpcUserIDRequiredError(RpcError):
    """Request must be sent with UserID field."""

    shv_error_code = RpcErrorCode.USER_ID_REQUIRED


class RpcNotImplementedError(RpcError):
    """Called method that is not implemented right now but valid."""

    shv_error_code = RpcErrorCode.NOT_IMPLEMENTED


RpcError.shv_error_map = {
    e.shv_error_code: e
    for e in (
        RpcInvalidRequestError,
        RpcMethodNotFoundError,
        RpcInvalidParamError,
        RpcInternalError,
        RpcParseError,
        RpcMethodCallTimeoutError,
        RpcMethodCallCancelledError,
        RpcMethodCallExceptionError,
        RpcLoginRequiredError,
        RpcUserIDRequiredError,
        RpcNotImplementedError,
    )
}



"""Common state for the RPC broker that works as server in the SHV network."""

from __future__ import annotations

import abc
import asyncio
import collections
import collections.abc
import contextlib
import logging
import secrets
import time
import typing

from ..rpcerrors import (
    RpcError,
    RpcInvalidParamError,
    RpcLoginRequiredError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
)
from ..rpclogin import RpcLogin
from ..rpcmessage import RpcMessage
from ..rpcmethod import RpcMethodAccess, RpcMethodDesc, RpcMethodFlags
from ..rpcparam import shvargt
from ..rpcri import rpcri_match, rpcri_relative_to
from ..rpctransport import RpcClient, RpcServer, create_rpc_server, init_rpc_client
from ..simplebase import SimpleBase
from ..simpleclient import SimpleClient
from ..value import SHVType
from .config import RpcBrokerConfigABC, RpcBrokerRoleABC
from .utils import nmax, nmin

logger = logging.getLogger(__name__)


class RpcBroker:
    """SHV RPC broker.

    The broker manages multiple RpcClient instances and exchanges messages
    between them.
    """

    class Client(SimpleBase):
        """Single client connected to the broker."""

        def __init__(
            self,
            client: RpcClient,
            broker: RpcBroker,
            *args: typing.Any,  # noqa ANN401
            **kwargs: typing.Any,  # noqa ANN401
        ) -> None:
            super().__init__(client, *args, **kwargs)
            self._send_queue: asyncio.Queue[RpcMessage] = asyncio.Queue(8)
            self.__broker = broker
            self.__broker_client_id: int | None = None
            self.__peer_is_broker: bool | None = None

        @property
        def broker(self) -> RpcBroker:
            """Broker this client is part of."""
            return self.__broker

        @property
        def broker_client_id(self) -> int | None:
            """Identifier assigned to the client in the broker."""
            return self.__broker_client_id

        @property
        def user_id(self) -> str:
            """Provide user ID for this Broker client."""
            # The minimal is the broker's name (child must extend this)
            return self.__broker.config.name

        @property
        @abc.abstractmethod
        def role(self) -> RpcBrokerRoleABC | None:
            """Role assigned in the broker.

            Can be used to query configuration for different rules for this
            client.
            """

        async def peer_is_broker(self) -> bool:
            """Identify if peer this client is handle for is broker."""
            if self.__peer_is_broker is None:
                self.__peer_is_broker = False
                with contextlib.suppress(RpcError):
                    lsr = await self.call("", "ls", ".broker")
                    self.__peer_is_broker = lsr if isinstance(lsr, bool) else False
            return self.__peer_is_broker

        def send(self, msg: RpcMessage) -> None:
            """Propagate given message to this peer.

            This provides a way for other clients to queue messages to be sent
            by this client. THis is because Broker peers are propagating
            messages from one to the other. This won't block but it also won't
            ensure that message won't be dropped.

            :param msg: Message to be sent.
            """
            try:
                self._send_queue.put_nowait(msg)
            except asyncio.QueueFull:
                logger.info("Message dropped due to overload: %s", str(msg))

        async def _send_loop(self) -> None:
            """Loop run in asyncio task to send queued messages."""
            with contextlib.suppress(EOFError):
                while True:
                    await self._send(await self._send_queue.get())
                    self._send_queue.task_done()

        async def _loop(self) -> None:
            send_task = asyncio.create_task(self._send_loop())
            try:
                await super()._loop()
            finally:
                send_task.cancel()
                with contextlib.suppress(asyncio.exceptions.CancelledError):
                    await send_task

        async def disconnect(self) -> None:  # noqa: D102
            logger.info("Disconnecting client with ID %s", str(self.broker_client_id))
            self._unregister()
            await super().disconnect()

        async def _message(self, msg: RpcMessage) -> None:
            assert self.role is not None

            if msg.is_request:
                # Set access granted to the level allowed by the role
                access = self.role.access_level(msg.path, msg.method)
                if access is None:
                    if msg.path not in {"", ".app", ".broker", ".broker/currentClient"}:
                        return await self._send(
                            msg.make_response(error=RpcMethodNotFoundError("No access"))
                        )
                    access = RpcMethodAccess.BROWSE
                # Limit access level in the message
                msg.rpc_access = (
                    access
                    if msg.rpc_access is None or msg.rpc_access > access
                    else msg.rpc_access
                )
                # Append user ID
                if msg.user_id is not None:
                    msg.user_id += (";" if msg.user_id else "") + self.user_id
                # Check if we should handle it ourself (else propagate it)
                if (cpath := self.__broker.client_on_path(msg.path)) is None:
                    return await super()._message(msg)
                # Protect sub-broker currentClient access
                if cpath[1] == ".broker/currentClient" and (
                    msg.method in {"subscribe", "unsubscribe"}
                    or msg.rpc_access < RpcMethodAccess.SUPER_SERVICE
                ):
                    return await self._send(
                        msg.make_response(error=RpcMethodNotFoundError("No access"))
                    )
                # Propagate to some peer
                assert self.__broker_client_id is not None
                msg.caller_ids = [*msg.caller_ids, self.__broker_client_id]
                msg.path = cpath[1]
                cpath[0].send(msg)

            elif msg.is_response:
                cids = list(msg.caller_ids)
                if not cids:  # no caller IDs means this is message for us
                    return await super()._message(msg)
                cid = cids.pop()
                msg.caller_ids = cids
                if (peer := self.__broker.get_client(cid)) is not None:
                    peer.send(msg)

            elif msg.is_signal:
                self.__broker.signal_from(msg, self)

        def _reset(self) -> None:
            self.__peer_is_broker = None
            self._unregister()

        def _register(self) -> None:
            if self.__broker_client_id is not None:
                self.__broker.unregister_client(self)
                self.__broker_client_id = None
            self.__broker_client_id = self.__broker.register_client(self)

        def _unregister(self) -> None:
            if self.__broker_client_id is not None:
                self.__broker.unregister_client(self)
                self.__broker_client_id = None

        def _ls(self, path: str) -> collections.abc.Iterator[str]:
            yield from super()._ls(path)
            if not path:
                yield ".broker"
            match path:
                case ".broker":
                    yield "currentClient"
                    yield "client"
                case ".broker/client":
                    yield from (
                        str(c.broker_client_id) for c in self.__broker.clients()
                    )
                case _:
                    for mnt, _ in self.broker.mounted_clients():
                        if not path:
                            yield mnt.split("/", maxsplit=1)[0]
                        elif mnt.startswith(path + "/"):
                            yield mnt[len(path) + 1 :].split("/", maxsplit=1)[0]

        def _dir(self, path: str) -> collections.abc.Iterator[RpcMethodDesc]:
            yield from super()._dir(path)
            match path:
                case ".broker":
                    yield RpcMethodDesc.getter(
                        "name",
                        "Null",
                        "String",
                        access=RpcMethodAccess.BROWSE,
                    )
                    yield RpcMethodDesc(
                        "clientInfo",
                        param="Int",
                        result="ClientInfo",
                        access=RpcMethodAccess.SUPER_SERVICE,
                    )
                    yield RpcMethodDesc(
                        "mountedClientInfo",
                        param="String",
                        result="ClientInfo",
                        access=RpcMethodAccess.SUPER_SERVICE,
                    )
                    yield RpcMethodDesc.getter(
                        "clients",
                        "Null",
                        "List[Int]",
                        access=RpcMethodAccess.SUPER_SERVICE,
                    )
                    yield RpcMethodDesc.getter(
                        "mounts",
                        "Null",
                        "List[String]",
                        access=RpcMethodAccess.SUPER_SERVICE,
                    )
                    yield RpcMethodDesc(
                        "disconnectClient",
                        param="Int",
                        access=RpcMethodAccess.SUPER_SERVICE,
                    )
                case ".broker/currentClient":
                    yield RpcMethodDesc(
                        "info", RpcMethodFlags.GETTER, result="ClientInfo"
                    )
                    yield RpcMethodDesc(
                        "subscribe",
                        param="String",
                        result="Bool",
                        access=RpcMethodAccess.BROWSE,
                    )
                    yield RpcMethodDesc(
                        "unsubscribe",
                        param="String",
                        result="Bool",
                        access=RpcMethodAccess.BROWSE,
                    )
                    yield RpcMethodDesc.getter(
                        "subscriptions",
                        result="List[String]",
                        access=RpcMethodAccess.BROWSE,
                    )

        async def _method_call(
            self,
            path: str,
            method: str,
            param: SHVType,
            access: RpcMethodAccess,
            user_id: str | None,
        ) -> SHVType:
            assert self.role is not None  # Otherwise handled in _message
            match path.split("/"):
                case [".broker"] if access >= RpcMethodAccess.SUPER_SERVICE:
                    match method:
                        case "name":
                            return self.__broker.config.name
                        case "clientInfo":
                            if not isinstance(param, int):
                                raise RpcInvalidParamError("Use Int")
                            client = self.broker.get_client(param)
                            return client.infomap() if client is not None else None
                        case "mountedClientInfo":
                            if not isinstance(param, str):
                                raise RpcInvalidParamError("Use String with SHV path")
                            client_pth = self.broker.client_on_path(param)
                            if client_pth is not None:
                                return client_pth[0].infomap()
                            return None
                        case "clients":
                            return list(
                                c.broker_client_id for c in self.broker.clients()
                            )
                        case "mounts":
                            return [mnt for mnt, _ in self.broker.mounted_clients()]
                        case "disconnectClient":
                            if not isinstance(param, int):
                                raise RpcInvalidParamError("Use Int")
                            client = self.broker.get_client(param)
                            if client is None:
                                raise RpcMethodCallExceptionError(
                                    f"No such client with ID: {param}"
                                )
                            await client.disconnect()
                            return None
                case [".broker", "currentClient"]:
                    match method:
                        case "info":
                            return self.infomap()
                        case "subscriptions":
                            return self._subscriptions()
                        case "subscribe":
                            sub = shvargt(param, 0, str)
                            ttl = shvargt(param, 1, int, -1)
                            return await self.__broker.subscribe(
                                self, sub, None if ttl < 0 else ttl
                            )
                        case "unsubscribe":
                            if not isinstance(param, str):
                                raise RpcInvalidParamError("Use string")
                            return await self.__broker.unsubscribe(self, param)
            return await super()._method_call(path, method, param, access, user_id)

        def infomap(self) -> dict[str, SHVType]:
            """Produce Map with client's info.

            This is info provided to administator to inform it about this
            client.
            """
            return {
                "clientId": self.broker_client_id,
                "mountPoint": self.__broker.client_mountpoint(self),
                "subscriptions": self._subscriptions(),
                "idleTime": int((time.monotonic() - self.client.last_receive) * 1000),
                "idleTimeMax": int(self.IDLE_TIMEOUT * 1000),
                "role": self.role.name if self.role is not None else None,
                "client": str(self.client),
            }

        def _subscriptions(self) -> dict[str, int | None]:
            now = time.time()
            return {
                str(ri): int(deadline - now) if deadline is not None else None
                for ri, deadline in self.__broker.subscriptions(self)
                if deadline is None or deadline > now
            }

    class LoginClient(Client):
        """Broker's client that expects login from client."""

        APP_NAME = "pyshvbroker"
        """Name reported as application name for pyshvbroker."""

        IDLE_TIMEOUT_LOGIN: float = 5
        """:attr:`IDLE_TIMEOUT` set for clients without user.

        This is intentionally shorter to quickly disconnect inactive clients
        that are not participating in SHV RPC.
        """

        def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa ANN204
            super().__init__(*args, **kwargs)
            self.IDLE_TIMEOUT = self.IDLE_TIMEOUT_LOGIN
            self._role: RpcBrokerRoleABC | None = None
            self._login: RpcLogin | None = None
            self._nonce: str = ""

        @property
        def role(self) -> RpcBrokerRoleABC | None:
            """Role of the user used to login to the broker."""
            return self._role

        @property
        def user_id(self) -> str:
            """Provide user ID for this Broker client."""
            s = super().user_id
            n = self._login.username if self._login else "?"
            return f"{n}{':' if s else ''}{s}"

        async def _loop(self) -> None:
            activity_task = asyncio.create_task(self._activity_loop())
            try:
                await super()._loop()
            finally:
                activity_task.cancel()
                with contextlib.suppress(asyncio.exceptions.CancelledError):
                    await activity_task

                self._unregister()
            self.client.disconnect()
            await self.client.wait_disconnect()

        async def _activity_loop(self) -> None:
            """Loop run alongside with :meth:`_loop`.

            It disconnects this client if it is idling.
            """
            while self.client.connected:
                t = time.monotonic() - self.client.last_receive
                if t < self.IDLE_TIMEOUT:
                    await asyncio.sleep(self.IDLE_TIMEOUT - t)
                else:
                    await self.disconnect()

        async def _message(self, msg: RpcMessage) -> None:
            # Login is required before we start handling messages as we would.
            if self._role is not None:
                return await super()._message(msg)
            if not msg.is_request:
                return  # Ignore anything other than requests before login
            if not msg.path:
                if msg.method == "hello":
                    if not self._nonce:
                        self._nonce = secrets.token_hex(5)  # ten characters
                    await self._send(msg.make_response({"nonce": self._nonce}))
                    return
                if self._nonce and msg.method == "login":
                    try:
                        self._login = RpcLogin.from_shv(msg.param)
                        self._role = self.broker.config.login(self._login, self._nonce)
                    except RpcError as exp:
                        await self._send(msg.make_response(error=exp))
                        return
                    if self._role is None:
                        error = RpcMethodCallExceptionError("Invalid login")
                        await self._send(msg.make_response(error=error))
                        return
                    try:
                        self._register()
                    except ValueError as exc:
                        error = RpcMethodCallExceptionError(str(exc))
                        await self._send(msg.make_response(error=error))
                        return
                    self.IDLE_TIMEOUT = float(
                        self._login.idle_timeout or type(self).IDLE_TIMEOUT
                    )
                    device_id = self._login.device_id
                    logger.info(
                        "Client %d logged in as user: %s%s",
                        self.broker_client_id,
                        self._login.username,
                        f" (deviceId={device_id})" if device_id else "",
                    )
                    await self._send(msg.make_response())
                    return
            await self._send(
                msg.make_response(
                    error=RpcLoginRequiredError(
                        "Use hello and login methods"
                        if self._nonce
                        else "Use hello method"
                    )
                )
            )

        def _reset(self) -> None:
            super()._reset()
            self._nonce = ""
            self._login = None
            self._role = None

        def infomap(self) -> dict[str, SHVType]:
            """Produce Map with client's info.

            This is info provided to administator to inform it about this
            client.

            On top of the standard one it provides also info about device ID.
            """
            assert self._login is not None  # Should not be called before login
            return {
                **super().infomap(),
                "userName": self._login.username,
                "deviceId": self._login.device_id,
            }

    class ConnectClient(SimpleClient, Client):
        """Broker client that activelly connects to some other peer."""

        APP_NAME = "pyshvbroker-client"
        """Name reported as application name for pyshvbroker connection."""

        def __init__(
            self,
            *args: typing.Any,  # noqa ANN204
            role: RpcBrokerRoleABC,
            **kwargs: typing.Any,  # noqa ANN204
        ) -> None:
            super().__init__(*args, **kwargs)
            self._role = role

        @property
        def role(self) -> RpcBrokerRoleABC:  # noqa D102
            return self._role

        async def _login(self) -> None:
            await super()._login()
            self._register()

    def __init__(self, config: RpcBrokerConfigABC) -> None:
        self.config = config
        """Configuration of the RPC Broker."""
        self.servers: dict[str, RpcServer] = {}
        """All servers managed by Broker where keys are their configured names."""
        self._clients: dict[int, RpcBroker.Client] = {}
        self._mounts: dict[str, int] = {}
        self._reserved_mount_points: set[str] = set()
        self._subs: dict[str, dict[int, float | None]] = {}
        """Subscriptions.

        RI filters signals and value contains map of all client IDs this
        subscription was requested by and specified end of the subscription
        live for that specific client.
        """
        self._subs_task: asyncio.Task | None = None
        self.__subs_changed = asyncio.Event()
        self.__next_caller_id = 0

    def register_client(self, client: Client) -> int:
        """Register RPC peer to the broker.

        :param client: Client object to register.
        :returns: Assigned client's ID in this broker.
        :raise ValueError: when registering with these parameters is for some
          reason not possible.
        """
        if client.role is None:
            raise ValueError("Client must have role assigned.")
        if client.broker_client_id is not None:
            raise ValueError("Client already registered.")
        # TODO try to reuse older caller IDs to send smaller messages but we
        # need to do it only after some given delay. We can save time of client
        # disconnect to the self.clients and use that to identify when we can
        # reuse caller id.
        cid = self.__next_caller_id
        self.__next_caller_id += 1
        self._clients[cid] = client
        mount_point: str | None = None
        if mount_point := client.role.mount_point(set(self._mounts)):
            mount_point = mount_point.rstrip("/")
            if mount_point.partition("/")[0] in {"", ".app", ".broker", ".device"}:
                raise ValueError("Mount point is not allowed")
            if any(
                mount_point == mnt
                or mnt.startswith(mount_point + "/")
                or mount_point.startswith(mnt + "/")
                for mnt in self._mounts
            ):
                raise ValueError("Mount point is already mounted")
            self.__lsmod_msg(mount_point, True)
            self._mounts[mount_point] = cid
        for ri in client.role.initial_subscriptions():
            self._subs.setdefault(ri, {})[cid] = None
        self._subs_notify()
        logger.info(
            "Client registered to broker with ID %d%s",
            cid,
            f" and mount point {mount_point}" if mount_point else "",
        )
        return cid

    def unregister_client(self, client: Client) -> None:
        """Unregister RPC peer and thus remove it from broker.

        :param client: Client object registration to be removed.
        :raise ValueError: when passed client is not registered.
        """
        cid = client.broker_client_id
        if cid is None:
            raise ValueError("Client not registered")
        if cid not in self._clients:
            raise ValueError("Client is not registered in this broker")
        if mount_point := self.client_mountpoint(client):
            del self._mounts[mount_point]
            self.__lsmod_msg(mount_point, False)
        for subs in self._subs.values():
            subs.pop(cid, None)
        self._subs = {n: v for n, v in self._subs.items() if v}
        del self._clients[cid]
        self._subs_notify()
        logger.info("Client with ID %d unregistered from broker", cid)

    def __lsmod_msg(self, path: str, new: bool) -> None:
        root = ""
        for node in path.split("/"):
            nroot = root + ("/" if root else "") + node
            if any(m.startswith(nroot) for m in self._mounts):
                root = nroot
            else:
                break
        self.signal(RpcMessage.signal(root, "lsmod", "ls", {node: new}))

    def get_client(self, cid: int | str) -> Client | None:
        """Lookup client with given ID.

        :param cid: ID of the client either as string or as integer.
        :return: :class:`RpcBroker.Client` when such client is located and ``None``
            otherwise.
        """
        try:
            return self._clients.get(cid if isinstance(cid, int) else int(cid, 10))
        except ValueError:
            return None

    def client_on_path(self, path: str) -> tuple[Client, str] | None:
        """Locate client mounted on given path.

        :return: client associated with this mount point and path relative to
            the client or None if there is no such client.
        """
        if path.startswith(".broker/client/"):
            pth = path.split("/")
            client = self.get_client(pth[2])
            return (client, "/".join(pth[3:])) if client else None

        # Note: we do not allow recursive mount points and thus first match is the
        # correct and the only client.
        for mnt, cid in self._mounts.items():
            if path.startswith(mnt + "/") or path == mnt:
                return self._clients[cid], path[len(mnt) + 1 :]
        return None

    def clients(self) -> collections.abc.Iterator[Client]:
        """Iterate over all clients participating in the broker."""
        yield from self._clients.values()

    def mounted_clients(self) -> collections.abc.Iterator[tuple[str, Client]]:
        """Goes through participating clients and provides those with mount point."""
        return ((mnt, self._clients[cid]) for mnt, cid in self._mounts.items())

    def client_mountpoint(self, client: Client) -> str | None:
        """Get mount point for this client.

        :param client: The client object that mount point should be received for.
        :return: The mount point of the client or ``None`` in case there is
          none.
        """
        try:
            return next(
                mnt
                for mnt, cid in self._mounts.items()
                if cid == client.broker_client_id
            )
        except StopIteration:
            return None

    def subscriptions(
        self, client: Client | None = None
    ) -> collections.abc.Iterator[tuple[str, float | None]]:
        """Iterate over subscriptions this broker manages.

        :param client: The optional filtering over client's subscriptions.
        :return: Iterator over subscriptions and their deadline (when they are
          no longer valid).
        """
        yield from (
            (
                s,
                v[client.broker_client_id]
                if client is not None and client.broker_client_id is not None
                else nmax(iter(v.values())),
            )
            for s, v in self._subs.items()
            if client is None or client.broker_client_id in v
        )

    async def subscribe(self, client: Client, ri: str, ttl: int | None = None) -> bool:
        """Add given subscription as being requested by given client.

        :param client: Client object this subscription is requested by.
        :param ri: The resource identifier for the subscription.
        :param ttl: Time to live for this subscription.
        :returns: ``False`` if this subscription was already present or ``True``
          if it is a new one.
        """
        if client.broker_client_id is None:
            raise ValueError("Client must be registered")
        subset = self._subs.setdefault(ri, {})
        res = client.broker_client_id not in subset
        subset[client.broker_client_id] = time.time() + ttl if ttl is not None else None
        self._subs_notify()
        return res

    async def unsubscribe(self, client: Client, ri: str) -> bool:
        """Remove given subscription as being requested by given client.

        :param client: Client object this subscription is requested by.
        :param ri: The resource identifier for the subscription.
        :returns: ``False`` if no subscription was removed and ``True``
          otherwise.
        """
        if client.broker_client_id is None:
            raise ValueError("Client must be registered")
        if ri not in self._subs:
            return False
        self._subs[ri].pop(client.broker_client_id, None)
        self._subs = {n: v for n, v in self._subs.items() if v}
        self._subs_notify()
        return True

    def signal(self, msg: RpcMessage) -> None:
        """Send signal to the broker's clients.

        :param msg: Signal message to be sent.
        """
        msgaccess = msg.rpc_access or RpcMethodAccess.READ
        cids: set[int] = set()
        for sub, clients in self._subs.items():
            if rpcri_match(sub, msg.path, msg.source, msg.signal_name):
                cids |= {c for c in clients if c is not None}
        for cid in cids:
            client = self._clients[cid]
            assert client.role is not None
            access = client.role.access_level(msg.path, msg.source)
            if access is not None and access >= msgaccess:
                client.send(msg)

    def signal_from(self, msg: RpcMessage, client: Client) -> None:
        """Send signal to the broker's client as comming from given client.

        :param msg: Signal message to be sent.
        :param client: Client ID or object signal should be propagated from.
        """
        if mnt := self.client_mountpoint(client):
            msg.path = mnt + ("/" if msg.path else "") + msg.path
            self.signal(msg)

    def _subs_notify(self) -> None:
        """Alert :meth:`_subs_loop` that changes to the subscriptions was made.

        This not only notifies running subscription loop but also starts it when
        it is not running.
        """
        if self._subs_task is None or self._subs_task.done():
            self._subs_task = asyncio.create_task(self._subs_loop())
        else:
            self.__subs_changed.set()

    async def _subs_loop(self) -> None:
        """Subscriptions management loop.

        This loop manages subscriptions in broker. It removes timed out
        subscriptions and propagates needed subscriptions to the sub-brokers.
        """
        loop = asyncio.get_running_loop()
        subsubs: dict[str, float] = {}
        subbrokers: set[int] = set()
        # TODO use heuristic to increase the delay for long term subscriptions
        while True:
            now = loop.time()
            # Remove any timed-out subscriptions
            for ri in self._subs:
                self._subs[ri] = {
                    k: v for k, v in self._subs[ri].items() if v is None or v > now
                }
            self._subs = {k: v for k, v in self._subs.items() if v}
            # Propagate subscriptions to subbrokers
            nsubbrokers: set[int] = set()
            for mnt, client in self.mounted_clients():
                # TODO we can't await in this loop to not modify the source
                # TODO this relies heavilly on SHV 3.0 and not all mounted
                # brokers will have that. Do we want to just ignore those or
                # do we want to implemented dedicated support for them?
                if not await client.peer_is_broker() or not await client.peer_is_shv3():
                    continue
                for ri in self._subs:
                    if (
                        client.broker_client_id not in subbrokers
                        or subsubs.get(ri, now) <= now
                    ) and (rri := rpcri_relative_to(ri, mnt)):
                        client.send(
                            RpcMessage.request(
                                ".broker/currentClient", "subscribe", [str(rri), 120]
                            )
                        )
                assert client.broker_client_id is not None
                nsubbrokers.add(client.broker_client_id)
            subbrokers = nsubbrokers
            subsubs = {
                ri: (now + 60)
                if ri not in subsubs or subsubs[ri] <= now
                else subsubs[ri]
                for ri in self._subs
            }
            # Now identify when we should do this again
            dlsub = nmin(*(iter(d.values()) for d in self._subs.values()))
            deadline = nmin(iter(subsubs.values()), dlsub)
            if dlsub is None and not subbrokers:
                return  # Terminate this task as it is not required
            # Now wait for event or deadline
            with contextlib.suppress(TimeoutError):
                async with asyncio.timeout_at(deadline):
                    await self.__subs_changed.wait()
                    self.__subs_changed.clear()

    async def start_serving(self, connect_timeout: float = 1.0) -> None:
        """Start accepting connections on all configured servers.

        This also initializes configured connections.

        :param connect_timeout: The timeout before we stop waiting for
           connections to be established.
        """

        def add(client: RpcClient) -> None:
            self.LoginClient(client, self)

        # We create our connect clients first to ensure that they get
        # preferential setup (and also preferential error detection).
        connect_clients = []
        for url, role in self.config.connections():
            connect_clients.append(
                self.ConnectClient(
                    init_rpc_client(url), url.login, self, role=role, reconnects=-1
                )
            )
            if mount_point := role.mount_point(self._reserved_mount_points):
                self._reserved_mount_points.add(mount_point)

        # Now start all servers so clients can connect
        for url in self.config.listens():
            surl = str(url)
            if surl not in self.servers:
                self.servers[surl] = await create_rpc_server(add, url)

        # Lastly wait specified time for clients to connect
        async with asyncio.timeout(connect_timeout):
            await asyncio.gather(
                *(client.wait_for_login() for client in connect_clients),
                return_exceptions=True,
            )

    async def serve_forever(self) -> None:
        """Serve all configured servers and block.

        This is a convenient coroutine that you can use to block main task until broker
        termination. You do not have to await this coroutine to run broker, awaiting on
        :meth:`start_serving` and keeping the loop running is enough to get broker
        working.
        """
        await self.start_serving()
        await asyncio.gather(
            *(server.wait_closed() for server in self.servers.values()),
            return_exceptions=True,
        )
        # TODO handle returned errors

    def close(self) -> None:
        """Request stop listening.

        This stops all servers and thus they no longer accept a new connections but the
        old connections are still kept and working.
        """
        for server in self.servers.values():
            server.close()

    async def wait_closed(self) -> None:
        """Wait for close to complete."""
        await asyncio.gather(
            *(server.wait_closed() for server in self.servers.values()),
            return_exceptions=True,
        )

    async def terminate(self) -> None:
        """Request termination of the broker.

        This closes broker as well as disconnects all established clients.
        """

        async def client_disconnect(client: RpcBroker.Client) -> None:
            try:
                await client.disconnect()
            except Exception as exc:
                logger.error("Disconnect failed", exc_info=exc)

        for server in self.servers.values():
            server.terminate()
        await asyncio.gather(
            *(client_disconnect(c) for c in self._clients.values()),
            *(s.wait_terminated() for s in self.servers.values()),
        )
        if self._subs_task is not None:
            if not self._subs_task.done():
                self._subs_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._subs_task



"""SHV RPC URL used to specify connection and listen settings."""

import dataclasses
import enum
import functools
import getpass
import urllib.parse

from .rpclogin import RpcLogin, RpcLoginType


@functools.lru_cache
def _get_user() -> str:
    # getuser fails if there is no account assigned to the UID in the system
    try:
        return getpass.getuser()
    except KeyError:
        return "none"


class RpcProtocol(enum.Enum):
    """Enum of supported RPC protocols by this Python implementation."""

    TCP = enum.auto()
    """TCP/IP with messages transported using Stream transport layer."""
    TCPS = enum.auto()
    """TCP/IP with messages transported using Serial transport layer."""
    SSL = enum.auto()
    """TLS TCP/IP with messages transported using Stream transport layer."""
    SSLS = enum.auto()
    """TLS TCP/IP with messages transported using Serial transport layer."""
    UNIX = enum.auto()
    """Unix local domain named socket using Stream transport layer."""
    UNIXS = enum.auto()
    """Unix local domain named socket using Reliable Serial transport layer."""
    TTY = enum.auto()
    """Serial transport layer."""
    WS = enum.auto()
    """WebSockets transport layer."""
    WSS = enum.auto()
    """WebSockets secure transport layer."""


@dataclasses.dataclass
class RpcUrl:
    """SHV RPC URL specifier.

    This is unified locator for SHV RPC connections that is used to specify
    connections. It is implemented as :func:`dataclasses.dataclass` and you
    can set properties directly or you can parse string URL.

    The URL format is defined in `SHV standard
    <https://silicon-heaven.github.io/shv-doc/rpcurl.html>`_.
    """

    # URL primary fields
    location: str
    """Hostname of the SHV RPC server or path to the socket."""
    port: int = -1
    """Port the SHV RPC server is listening on."""
    protocol: RpcProtocol = RpcProtocol.TCP
    """SHV RPC protocol used to communicate (This is scheme in URL therminology)"""
    login: RpcLogin = dataclasses.field(default_factory=RpcLogin)
    """Parameters for RPC Broker login."""

    # TTY
    baudrate: int = 115200
    """Baudrate used for some of the link protocols."""

    def __post_init__(self) -> None:
        """Deduce the correct value for port."""
        if self.port == type(self).port:
            self.port = {
                RpcProtocol.TCP: 3755,
                RpcProtocol.TCPS: 3765,
                RpcProtocol.SSL: 3756,
                RpcProtocol.SSLS: 3766,
            }.get(self.protocol, self.port)

    def __str__(self) -> str:
        return self.to_url()

    @classmethod
    def parse(cls, url: str) -> "RpcUrl":
        """Parse string URL to the object representation.

        :param url: URL in string format.
        :return: New :class:`RpcUrl` instance.
        :raise ValueError: when invalid URL is passed.
        """
        sr = urllib.parse.urlsplit(url, scheme="unix", allow_fragments=False)
        pqs = urllib.parse.parse_qs(sr.query)

        protocols = {
            "tcp": RpcProtocol.TCP,
            "tcps": RpcProtocol.TCPS,
            "ssl": RpcProtocol.SSL,
            "ssls": RpcProtocol.SSLS,
            "unix": RpcProtocol.UNIX,
            "unixs": RpcProtocol.UNIXS,
            "serial": RpcProtocol.TTY,
            "tty": RpcProtocol.TTY,
            "ws": RpcProtocol.WS,
            "wss": RpcProtocol.WSS,
        }
        if sr.scheme not in protocols:
            raise ValueError(f"Invalid scheme: {sr.scheme}")
        protocol = protocols[sr.scheme]

        res = cls("", protocol=protocol)

        res.login.username = sr.username or res.login.username
        if protocol in {
            RpcProtocol.TCP,
            RpcProtocol.TCPS,
            RpcProtocol.SSL,
            RpcProtocol.SSLS,
        }:
            res.location = sr.hostname or ""
            if sr.port is not None:
                res.port = int(sr.port)
            if sr.path:
                raise ValueError(f"Path is not supported for {sr.scheme}: {sr.path}")
        elif protocol in {RpcProtocol.UNIX, RpcProtocol.UNIXS, RpcProtocol.TTY}:
            res.location = f"/{sr.netloc}{sr.path}" if sr.netloc else sr.path
        elif protocol in {RpcProtocol.WS, RpcProtocol.WSS}:
            res.location = f"{sr.netloc}{sr.path}"
        else:
            raise NotImplementedError  # pragma: no cover

        if opts := pqs.pop("user", []):
            res.login.username = opts[0]
        if opts := pqs.pop("shapass", []):
            if len(opts[0]) != 40:
                raise ValueError("SHA1 password must have 40 characters.")
            res.login.password = opts[0]
            res.login.login_type = RpcLoginType.SHA1
            # We prefer SHA1 password and thus discard plain if both are present
            pqs.pop("password", [])
        elif opts := pqs.pop("password", []):
            res.login.password = opts[0]
            res.login.login_type = RpcLoginType.PLAIN
        if opts := pqs.pop("devid", []):
            res.login.device_id = opts[0]
        if opts := pqs.pop("devmount", []):
            res.login.device_mount_point = opts[0]
        if protocol is RpcProtocol.TTY:
            if opts := pqs.pop("baudrate", []):
                res.baudrate = int(opts[0])

        if pqs:
            raise ValueError(f"Unsupported URL queries: {', '.join(pqs.keys())}")

        return res

    def to_url(self, public: bool = False) -> str:
        """Convert to string URL.

        :param public: You can pass ``True`` to not include login credentials.
        :returns: The string representation of the RPC URL.
        """
        protocols = {
            RpcProtocol.TCP: "tcp",
            RpcProtocol.TCPS: "tcps",
            RpcProtocol.SSL: "ssl",
            RpcProtocol.SSLS: "ssls",
            RpcProtocol.UNIX: "unix",
            RpcProtocol.UNIXS: "unixs",
            RpcProtocol.TTY: "serial",
            RpcProtocol.WS: "ws",
            RpcProtocol.WSS: "wss",
        }
        user_added = not self.login.username or self.login.username == RpcLogin.username
        if self.protocol in {
            RpcProtocol.TCP,
            RpcProtocol.TCPS,
            RpcProtocol.SSL,
            RpcProtocol.SSLS,
        }:
            netloc = "//"
            if not user_added:
                netloc += f"{self.login.username}@"
                user_added = True
            if ":" in self.location:
                netloc += f"[{self.location}]"
            else:
                netloc += self.location
            netloc += f":{self.port}"
        elif self.protocol in {
            RpcProtocol.UNIX,
            RpcProtocol.UNIXS,
            RpcProtocol.TTY,
            RpcProtocol.WS,
            RpcProtocol.WSS,
        }:
            netloc = self.location
        else:
            raise NotImplementedError  # pragma: no cover

        opts: list[str] = []
        if not user_added:
            opts.append(f"user={urllib.parse.quote(self.login.username)}")
            user_added = True
        if self.login.device_id:
            opts.append(f"devid={urllib.parse.quote(self.login.device_id)}")
        if self.login.device_mount_point:
            opts.append(f"devmount={urllib.parse.quote(self.login.device_mount_point)}")
        if self.login.password and not public:
            if self.login.login_type is RpcLoginType.SHA1:
                opts.append(f"shapass={self.login.password}")
            elif self.login.login_type is RpcLoginType.PLAIN:
                opts.append(f"password={urllib.parse.quote(self.login.password)}")
            else:
                raise NotImplementedError()  # pragma: no cover
        if self.baudrate != type(self).baudrate:
            opts.append(f"baudrate={self.baudrate}")

        return (
            f"{protocols[self.protocol]}:{netloc}{'?' if opts else ''}{'&'.join(opts)}"
        )



"""Check that our tools for working with values are correct."""

import pytest

from shv import SHVGetKey, shvarg, shvget


@pytest.mark.parametrize(
    "value,key,default,expected",
    (
        (None, "foo", 3, 3),
        ({"foo": 42}, "foo", 3, 42),
        ({}, "foo", 3.0, 3.0),
        ({"foo": {"foo": 42}}, ("foo", "foo"), 3, 42),
        ({"bar": "foo"}, ("foo", "foo"), 3, 3),
        ({"foo": {"bar": 33}}, (SHVGetKey("foo", 2), SHVGetKey("bar", 3)), 3, 33),
        ({2: {3: 33}}, (SHVGetKey("foo", 2), SHVGetKey("bar", 3)), 3, 33),
        (None, (), 3, 3),
        (42, (), 3, 42),
    ),
)
def test_shvget(value, key, default, expected):
    """Check implementation of shvget."""
    assert shvget(value, key, default) == expected


@pytest.mark.parametrize(
    "value,index,default,expected",
    (
        (None, 0, 3, 3),
        (41, 1, 4, 4),
        (41, 0, 4, 41),
        ([41, 42, 43], 1, 3, 42),
        ([41, None, 43], 1, 3, 3),
        ([], 0, 3, 3),
        ([41], 1, 3, 3),
        ([41], 0, 3, 41),
        ("foo", 1, None, None),
        ("foo", 0, 4, "foo"),
    ),
)
def test_shvarg(value, index, default, expected):
    """Check implementation of shvarg."""
    assert shvarg(value, index, default) == expected



"""Check implementation of :class:`RpcLogin`."""

import re

import pytest

from shv import RpcInvalidParamError, RpcLogin, RpcLoginType


def test_options_init():
    assert RpcLogin(opt_device_id="foo", opt_device_mount_point="test/foo").options == {
        "device": {"deviceId": "foo", "mountPoint": "test/foo"}
    }


def test_device_id():
    obj = RpcLogin()
    assert obj.device_id is None
    obj.device_id = "foo"
    assert obj.device_id == "foo"
    assert obj.options == {"device": {"deviceId": "foo"}}
    obj.device_id = None
    assert obj.device_id is None
    assert obj.options == {}


def test_device_mount_point():
    obj = RpcLogin(opt_device_id="foo")
    assert obj.device_mount_point is None
    obj.device_mount_point = "test/foo"
    assert obj.device_mount_point == "test/foo"
    assert obj.options == {"device": {"deviceId": "foo", "mountPoint": "test/foo"}}
    obj.device_mount_point = None
    assert obj.device_mount_point is None
    assert obj.options == {"device": {"deviceId": "foo"}}


def test_idle_timeout():
    obj = RpcLogin(opt_device_id="foo")
    assert obj.idle_timeout is None
    obj.idle_timeout = 300
    assert obj.idle_timeout == 300
    assert obj.options == {"device": {"deviceId": "foo"}, "idleWatchDogTimeOut": 300}
    obj.idle_timeout = None
    assert obj.idle_timeout is None
    assert obj.options == {"device": {"deviceId": "foo"}}


def test_extend_options():
    obj = RpcLogin(options={"foo": 42, "ord": {"left": "<"}, "wtf": {"this": 1}})
    obj.extend_options({"foo": 24, "ord": {"right": ">"}, "wtf": {"this": None}})
    assert obj.options == {"foo": 24, "ord": {"left": "<", "right": ">"}}


@pytest.mark.parametrize(
    "obj,password,login_type,expected",
    (
        (RpcLogin("admin"), "", RpcLoginType.PLAIN, True),
        (RpcLogin("admin"), "invalid", RpcLoginType.PLAIN, False),
        (
            RpcLogin(
                "admin", "a7162d463b28666737f63034db39f03bca59b060", RpcLoginType.SHA1
            ),
            "",
            RpcLoginType.PLAIN,
            True,
        ),
        (
            RpcLogin(
                "admin", "a7162d463b28666737f63034db39f03bca59b060", RpcLoginType.SHA1
            ),
            "invalid",
            RpcLoginType.PLAIN,
            False,
        ),
        (
            RpcLogin("admin"),
            "da39a3ee5e6b4b0d3255bfef95601890afd80709",
            RpcLoginType.SHA1,
            True,
        ),
        (
            RpcLogin("admin"),
            "81f344a7686a80b4c5293e8fdc0b0160c82c06a8",
            RpcLoginType.SHA1,
            False,
        ),
        (
            RpcLogin(
                "admin", "a7162d463b28666737f63034db39f03bca59b060", RpcLoginType.SHA1
            ),
            "81f344a7686a80b4c5293e8fdc0b0160c82c06a8",
            RpcLoginType.SHA1,
            False,
        ),
    ),
)
def test_validate_password(obj, password, login_type, expected):
    assert obj.validate_password(password, "nonce", login_type) is expected


@pytest.mark.parametrize(
    "obj,custom_options,trusted,shv",
    (
        (
            RpcLogin("admin", "admin!123"),
            None,
            True,
            {
                "login": {"password": "admin!123", "user": "admin", "type": "PLAIN"},
                "options": {},
            },
        ),
        (
            RpcLogin("admin", options={"device": {"mountPoint": "test/this"}}),
            {"other": 42, "device": {"deviceId": "foo"}},
            False,
            {
                "login": {
                    "password": "a7162d463b28666737f63034db39f03bca59b060",
                    "user": "admin",
                    "type": "SHA1",
                },
                "options": {
                    "device": {"deviceId": "foo", "mountPoint": "test/this"},
                    "other": 42,
                },
            },
        ),
    ),
)
def test_to_shv(obj, custom_options, trusted, shv):
    assert obj.to_shv("nonce", custom_options, trusted) == shv


@pytest.mark.parametrize(
    "shv,obj",
    (
        ({}, RpcLogin("")),
        (
            {"login": {"password": "", "user": "admin", "type": "PLAIN"}},
            RpcLogin("admin"),
        ),
        (
            {
                "login": {
                    "password": "a7162d463b28666737f63034db39f03bca59b060",
                    "user": "admin",
                    "type": "SHA1",
                },
                "options": {
                    "device": {"mountPoint": "test/this"},
                    "other": 42,
                },
            },
            RpcLogin(
                "admin",
                "a7162d463b28666737f63034db39f03bca59b060",
                RpcLoginType.SHA1,
                {
                    "device": {"mountPoint": "test/this"},
                    "other": 42,
                },
            ),
        ),
    ),
)
def test_from_shv(shv, obj):
    assert RpcLogin.from_shv(shv) == obj


@pytest.mark.parametrize(
    "shv,msg",
    (([], "Expected Map."),),
)
def test_from_shv_invalid(shv, msg):
    with pytest.raises(RpcInvalidParamError, match=re.escape(msg)):
        RpcLogin.from_shv(shv)



"""Connection over CAN bus."""

import logging

import can

from .abc import RpcClient, RpcServer

logger = logging.getLogger(__name__)


class RpcClientCAN(RpcClient):
    """RPC connection to some SHV peer over CAN bus.

    :param bus: CAN bus to be used for the communication.
    :param local: The address to be used as the local one for this peer.
    :param dest: The address of the destination peer.
    """

    def __init__(self, interface: str, local: int, dest: int) -> None:
        self.interface = interface
        """CAN interfaces to be used."""
        self.local_address = local
        """Address used for this local peer's end (1-255)."""
        self.dest_address = dest
        """Address of the peer this connection should be connected to."""
        self._bus: can.Bus | None = None
        self._reader: can.AsyncBufferedReader
        self._notifier: can.Notifier

    def __str__(self) -> str:
        return f"can:{self.interface}:{self.local_address}:{self.dest_address}"

    async def _send(self, msg: bytes) -> None:
        if self._bus is None:
            raise EOFError("Not connected")
        self._bus.send(msg)
        # TODO exception?

    async def _receive(self) -> bytes:
        if self._bus is None:
            raise EOFError("Not connected")
        msg = await reader.get_message()
        # TODO we need to manage multiple messages

    async def reset(self) -> None:
        """Reset or establish the connection.

        This method not only resets the existing connection but primarilly
        estableshes the new one.
        """
        if not self.connected:
            self._bus = can.Bus(interface=self.interface, receive_own_messages=False)
            self._reader = can.AsyncBufferedReader()
            self._notifier = can.Notifier(self._bus, [self._reader])
            logger.debug("%s: Connected", self)
        else:
            await super().reset()

    @property
    def connected(self) -> bool:
        """Check if client is still connected."""
        return self._bus is not None and self._bus.state is not can.BusState.ERROR

    def _disconnect(self) -> None:
        if self._bus is not None:
            self._bus.shutdown()

    async def wait_disconnect(self) -> None:
        """Wait for the client's disconnection."""
        pass  # Nothing to wait for


class RpcServerCAN(RpcServer):
    """RPC server listening on CAN bus."""

    def __init__(self, bus: can.Bus, address: int) -> None:
        self.bus = bus
        self._address = address



"""Common base for RPC clients and servers."""

import abc
import enum
import logging
import time
import typing

from ..chainpack import ChainPack
from ..rpcmessage import RpcMessage
from ..value import SHVIMap

logger = logging.getLogger(__name__)


class RpcClient(abc.ABC):
    """RPC connection to some SHV peer."""

    class Control(enum.Enum):
        """Control message that is received instead of :class:`RpcMessage`."""

        RESET = enum.auto()

    def __init__(self) -> None:
        self.last_send = time.monotonic()
        """Monotonic time when last message was sent on this connection.

        The initial value is time of the RpcClient creation.
        """
        self.last_receive = time.monotonic()
        """Monotonic time when last message was received on this connection.

        The initial value is time of the RpcClient creation.
        """

    @classmethod
    async def connect(cls, *args: typing.Any, **kwargs: typing.Any) -> typing.Self:  # noqa ANN401
        """Connect client.

        This conveniently combines object initialization and call to
        :meth:`reset`. All arguments are passed to the object initialization.
        """
        res = cls(*args, **kwargs)
        await res.reset()
        return res

    async def send(self, msg: RpcMessage) -> None:
        """Send the given SHV RPC Message.

        :param msg: Message to be sent
        :raise EOFError: when client is not connected.
        """
        await self._send(bytearray((ChainPack.ProtocolType,)) + msg.to_chainpack())
        self.last_send = time.monotonic()
        logger.debug("%s => %s", str(self), msg.to_string())

    @abc.abstractmethod
    async def _send(self, msg: bytes) -> None:
        """Child's implementation of message sending."""

    async def receive(self, raise_error: bool = True) -> RpcMessage | Control:
        """Read next received RPC message or wait for next to be received.

        :param raise_error: If RpcError should be raised or not.
        :return: Next RPC message is returned or :class:`Control` for special control
          messages.
        :raise RpcError: When mesasge is error and ``raise_error`` is `True`.
        :raise EOFError: in case EOF is encountered.
        """
        while True:
            data = await self._receive()
            self.last_receive = time.monotonic()
            if len(data) > 1:
                if data[0] == ChainPack.ProtocolType:
                    try:
                        shvdata = ChainPack.unpack(data[1:])
                    except ValueError:
                        pass
                    else:
                        if isinstance(shvdata, SHVIMap):
                            msg = RpcMessage(shvdata)
                            logger.debug("%s <= %s", self, msg.to_string())
                            if raise_error and msg.is_error:
                                raise msg.rpc_error
                            return msg
            elif len(data) == 1 and data[0] == 0:
                logger.debug("%s <= Control message RESET", self)
                return self.Control.RESET
            logger.debug("%s <= Invalid message received: %s", self, data)

    @abc.abstractmethod
    async def _receive(self) -> bytes:
        """Message receive implementation.

        :return: bytes of received message (complete valid message).
        :raise EOFError: if end of the connection is encountered.
        """

    async def reset(self) -> None:
        """Reset the connection.

        This sends reset to the peer and thus it is instructed to forget anything that
        might have been associated with this client.

        This can also try reconnect if client supports it.

        This can raise not only :class:`EOFError` but also other exception based
        on the client implementation.

        :raise EOFError: if peer is not connected and reconnect is not either supported
          or possible.
        """
        await self._send(bytes((0,)))
        logger.debug("%s => Control message RESET", self)

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        """Check if client is still connected.

        This is only local check. There is no communication done and thus depending on
        the transport layer this can be pretty much a lie. The only reliable tests is
        sending request and receiving respond to it.

        :return: ``True`` if client might still connected and ``False`` if we know that
            it is not.
        """

    def disconnect(self) -> None:
        """Close the connection."""
        if self.connected:
            logger.debug("%s: Disconnecting", self)
        self._disconnect()

    @abc.abstractmethod
    def _disconnect(self) -> None:
        """Child's implementation of message sending."""

    async def wait_disconnect(self) -> None:  # noqa D027
        """Wait for the client's disconnection."""


class RpcServer(abc.ABC):
    """RPC server listening for new SHV connections."""

    @abc.abstractmethod
    def is_serving(self) -> bool:
        """Check if server is accepting new SHV connections.

        :return: ``True`` if new connection can be accepted and ``False``
          otherwise.
        """

    @abc.abstractmethod
    async def listen(self) -> None:
        """Start accepting new SHV connections."""

    @abc.abstractmethod
    async def listen_forewer(self) -> None:
        """Listen and block the calling coroutine until cancelation."""

    @abc.abstractmethod
    def close(self) -> None:
        """Stop accepting new SHV connections."""

    @abc.abstractmethod
    async def wait_closed(self) -> None:
        """Wait for the server closure."""

    @abc.abstractmethod
    def terminate(self) -> None:
        """Terminate the server.

        This disconnects all accepted clients as well as stops accepting new SHV
        connections.
        """

    @abc.abstractmethod
    async def wait_terminated(self) -> None:
        """Wait for the server termination."""



"""Implementation of SHV RPC broker."""

from .broker import RpcBroker
from .config import RpcBrokerConfig
from .configabc import RpcBrokerConfigABC, RpcBrokerRoleABC

__all__ = [
    "RpcBroker",
    "RpcBrokerConfig",
    "RpcBrokerConfigABC",
    "RpcBrokerRoleABC",
]



"""Values used in SHV communication."""

import abc
import collections.abc
import datetime
import decimal
import functools
import itertools
import typing

SHVNullType: typing.TypeAlias = typing.Union[None, "SHVNull"]
SHVBoolType: typing.TypeAlias = typing.Union[bool, "SHVBool"]
SHVListType: typing.TypeAlias = collections.abc.Sequence["SHVType"]
SHVMapType: typing.TypeAlias = collections.abc.Mapping[str, "SHVType"]
SHVIMapType: typing.TypeAlias = collections.abc.Mapping[int, "SHVType"]
SHVType: typing.TypeAlias = (
    SHVNullType
    | SHVBoolType
    | int
    | float
    | decimal.Decimal
    | bytes
    | str
    | datetime.datetime
    | SHVListType
    | SHVMapType
    | SHVIMapType
    | SHVIMapType
    | SHVMapType
)
SHVMetaType: typing.TypeAlias = collections.abc.MutableMapping[int | str, SHVType]


class SHVMeta(abc.ABC):  # noqa B024
    """SHV values can have meta with attributes associated with them.

    This provides meta attribute that can be added to other types. Only create
    class that has as parent this class and the class you want to wrap. The
    SHVMeta provides you with meta attribute that you can use to store and
    access meta attributes. Thanks to it being both parent of SHVMeta as well as
    original type you can use isinstance to detect type. For example to check if
    type is `int` you can use ``isinstance(foo, int)``.

    .. warning::
        ``None`` and bool can't have meta assigned to the them just by simple
        inheritance and thus we define our custom types. Make sure that you
        always expect that you can get instance of :class:`SHVNull` instead of
        ``None`` and :class:`SHVBool` instead of ``True`` and ``False``.
    .. note::
        Meta is not intentionally included in the plain comparison to ensure
        that standard hashable types are still hashable while meta is
        modifiable. You can use :func:`shvmeta_eq` to compare with meta.
    """

    @property
    def meta(self) -> SHVMetaType:
        """Meta attributes for this SHV type."""
        if not hasattr(self, "_meta"):
            self._meta: SHVMetaType = {}
        return self._meta

    @staticmethod
    def new(value: object, meta: SHVMetaType | None = None) -> SHVType:
        """Create new value with given meta.

        This select an appropriate class based on the value passed.

        You can also use it to set meta value to newly created object because
        when `value` is of `SHVMeta` it returns the same object and only updates
        provided meta.
        """
        res: SHVMeta
        if isinstance(value, SHVMeta):
            res = value
        elif value is None:
            res = SHVNull()
        elif isinstance(value, bool):
            res = SHVBool(value)
        elif isinstance(value, int):
            res = SHVInt(value)
        elif isinstance(value, float):
            res = SHVFloat(value)
        elif isinstance(value, bytes):
            res = SHVBytes(value)
        elif isinstance(value, str):
            res = SHVStr(value)
        elif isinstance(value, datetime.datetime):
            res = SHVDatetime.fromtimestamp(value.timestamp(), value.tzinfo)
        elif isinstance(value, decimal.Decimal):
            res = SHVDecimal(value)
        elif isinstance(value, collections.abc.Sequence):
            res = SHVList(value)
        elif is_shvimap(value):
            res = SHVIMap(value)
        elif is_shvmap(value):
            res = SHVMap(value)
        else:
            raise ValueError(f"Invalid SHV value: {value!r}")
        if meta:
            res.meta.update(meta)
        return typing.cast(SHVType, res)


def shvmeta(value: object) -> SHVMetaType:
    """Get SHV Meta or provide empty dict as a fallback."""
    if isinstance(value, SHVMeta):
        return value.meta
    return {}


def shvmeta_eq(v1: object, v2: object) -> bool:
    """Perform comparison including the :class:`SHVMeta` not just plain values."""
    if shvmeta(v1) != shvmeta(v2):
        return False
    if isinstance(v1, SHVUInt) != isinstance(v2, SHVUInt):
        return False
    if (
        isinstance(v1, collections.abc.Sequence)
        and isinstance(v2, collections.abc.Sequence)
        and not (isinstance(v1, str | bytes) or isinstance(v2, str | bytes))
    ):
        return len(v1) == len(v2) and all(
            shvmeta_eq(v1[i], v2[i]) for i in range(len(v1))
        )
    if isinstance(v1, collections.abc.Mapping) and isinstance(
        v2, collections.abc.Mapping
    ):
        return all(
            k in v1 and k in v2 and shvmeta_eq(v1[k], v2[k])
            for k in set(itertools.chain(v1.keys(), v2.keys()))
        )
    return bool(v1 == v2)


class SHVNull(SHVMeta):
    """Null (None) with :class:`SHVMeta`."""

    def __bool__(self) -> bool:
        return False

    def __eq__(self, value: object) -> bool:
        return value is None or isinstance(value, SHVNull)

    def __hash__(self) -> int:
        return hash(None)


def is_shvnull(value: object) -> typing.TypeGuard[SHVNullType]:
    """Validate type of the value as either ``None`` or :class:`SHVNull`."""
    return value is None or isinstance(value, SHVNull)


class SHVBool(SHVMeta):
    """Boolean with :class:`SHVMeta`."""

    def __init__(self, value: bool) -> None:
        self._value = value

    def __bool__(self) -> bool:
        return self._value

    def __eq__(self, value: object) -> bool:
        return bool(value) is self._value

    def __hash__(self) -> int:
        return hash(self._value)


def is_shvbool(value: object) -> typing.TypeGuard[bool | SHVBool]:
    """Validate type of value as either :class:`bool` or :class:`SHVBool`."""
    return isinstance(value, bool | SHVBool)


class SHVInt(int, SHVMeta):
    """Integer with class:`SHVMeta`."""


class SHVUInt(int, SHVMeta):
    """Unsigned integer with :class:`SHVMeta`.

    There is no unsigned type in Python and thus compared to :class:`SHVInt`
    (that can be interchanged with `int`) you have to always use this class to
    represent unsigned integer.
    """


class SHVFloat(float, SHVMeta):
    """Float with :class:`SHVMeta`."""


class SHVDecimal(decimal.Decimal, SHVMeta):
    """Decimal with :class:`SHVMeta`."""


def decimal_rexp(value: decimal.Decimal) -> tuple[int, int]:
    """Decomposes decimal number to the mantissa and exponent.

    Python's Decimal does not directly provides a way to get mantissa and exponent. In
    SHV it is the preferred way to store decimal numbers and this this functions is
    provided for that.

    :param value: Decimal number to be decomposed to mantissa and exponent.
    :returns: Tuple with mantissa and decimal exponent.
    """
    t = value.as_tuple()
    mantissa = functools.reduce(
        lambda a, b: a + (b[1] * 10 ** b[0]), enumerate(reversed(t.digits)), 0
    ) * (-1 if t.sign else 1)
    assert isinstance(t.exponent, int)
    return mantissa, t.exponent


class SHVBytes(bytes, SHVMeta):
    """Bytes with :class:`SHVMeta`."""


class SHVStr(str, SHVMeta):
    """String with :class:`SHVMeta`."""


class SHVDatetime(datetime.datetime, SHVMeta):
    """Date and time with :class:`SHVMeta`."""


class SHVList(list[SHVType], SHVMeta):
    """List of :class:`SHVMeta` values."""


class SHVMap(dict[str, SHVType], SHVMeta):
    """Dictionary with :class:`SHVMeta`."""


class SHVIMap(dict[int, SHVType], SHVMeta):
    """Dictionary with :class:`SHVMeta`."""


def is_shvlist(value: object) -> typing.TypeGuard[SHVListType]:
    """Check if given value can be SHV List."""
    return (
        isinstance(value, collections.abc.Sequence)
        and not isinstance(value, str | bytes)
        and all(is_shvtype(v) for v in value)
    )


def is_shvmap(value: object) -> typing.TypeGuard[SHVMapType]:
    """Check if given value can be SHV Map."""
    return isinstance(value, collections.abc.Mapping) and all(
        isinstance(k, str) and is_shvtype(v) for k, v in value.items()
    )


def is_shvimap(value: object) -> typing.TypeGuard[SHVIMapType]:
    """Check if given value can be SHV IMap."""
    return isinstance(value, collections.abc.Mapping) and all(
        isinstance(k, int) and is_shvtype(v) for k, v in value.items()
    )


def is_shvtype(value: object) -> typing.TypeGuard[SHVType]:
    """Validate type of the value as SHVType."""
    return (
        is_shvnull(value)
        or is_shvbool(value)
        or isinstance(
            value, int | float | decimal.Decimal | bytes | str | datetime.datetime
        )
        or is_shvlist(value)
        or is_shvimap(value)
        or is_shvmap(value)
    )



"""Check implementation of SimpleClient."""

import asyncio
import dataclasses

import pytest

from shv import (
    RpcLoginType,
    RpcMethodAccess,
    RpcMethodDesc,
    RpcMethodFlags,
    SimpleClient,
    shvmeta,
)


@pytest.mark.parametrize(
    "path,method,params,result",
    (
        (".app", "ping", None, None),
        ("", "ls", None, [".app", ".broker"]),
        (
            "",
            "dir",
            None,
            [
                {
                    1: "dir",
                    2: 0,
                    3: "idir",
                    4: "odir",
                    5: 1,
                },
                {
                    1: "ls",
                    2: 0,
                    3: "ils",
                    4: "ols",
                    5: 1,
                    6: {"lsmod": "olsmod"},
                },
            ],
        ),
    ),
)
async def test_call(client, path, method, params, result):
    """Check that we can call various methods using blocking call."""
    res = await client.call(path, method, params)
    assert res == result
    assert shvmeta(res) == shvmeta(result)


@pytest.mark.parametrize(
    "path,result",
    (
        ("", [".app", ".broker"]),
        (".app", []),
        (".broker", ["currentClient", "client"]),
    ),
)
async def test_ls(client, path, result):
    """Verify that we can use ls method."""
    assert await client.ls(path) == result


@pytest.mark.parametrize(
    "path,name,result",
    (
        ("", ".broker", True),
        ("", "invalid", False),
        (".broker", "currentClient", True),
        (".broker", "foo", False),
    ),
)
async def test_ls_has_child(client, path, name, result):
    """Verify that child existence check works."""
    assert await client.ls_has_child(path, name) == result


@pytest.mark.parametrize(
    "path,result",
    (
        (
            "",
            [RpcMethodDesc.stddir(), RpcMethodDesc.stdls()],
        ),
        (
            ".broker",
            [
                RpcMethodDesc.stddir(),
                RpcMethodDesc.stdls(),
                RpcMethodDesc.getter(
                    "name", "Null", "String", access=RpcMethodAccess.BROWSE
                ),
                RpcMethodDesc(
                    "clientInfo",
                    param="Int",
                    result="ClientInfo",
                    access=RpcMethodAccess.SUPER_SERVICE,
                ),
                RpcMethodDesc(
                    "mountedClientInfo",
                    param="String",
                    result="ClientInfo",
                    access=RpcMethodAccess.SUPER_SERVICE,
                ),
                RpcMethodDesc.getter(
                    "clients", "Null", "List[Int]", access=RpcMethodAccess.SUPER_SERVICE
                ),
                RpcMethodDesc.getter(
                    "mounts",
                    "Null",
                    "List[String]",
                    access=RpcMethodAccess.SUPER_SERVICE,
                ),
                RpcMethodDesc(
                    "disconnectClient",
                    param="Int",
                    access=RpcMethodAccess.SUPER_SERVICE,
                ),
            ],
        ),
        (
            ".broker/currentClient",
            [
                RpcMethodDesc.stddir(),
                RpcMethodDesc.stdls(),
                RpcMethodDesc("info", RpcMethodFlags.GETTER, result="ClientInfo"),
                RpcMethodDesc("subscribe", param="String", result="Bool"),
                RpcMethodDesc("unsubscribe", param="String", result="Bool"),
                RpcMethodDesc.getter(
                    "subscriptions",
                    result="List[String]",
                    access=RpcMethodAccess.BROWSE,
                ),
            ],
        ),
    ),
)
async def test_dir(client, path, result):
    """Verify that we can use dir method."""
    res = await client.dir(path)
    assert res == result


@pytest.mark.parametrize(
    "path,name,result",
    (
        (".app", "ping", True),
        (".app", "invalid", False),
        (".broker/currentClient", "info", True),
    ),
)
async def test_dir_exists(client, path, name, result):
    """Verify that method existence check works."""
    res = await client.dir_exists(path, name)
    assert res == result


async def test_sha_login(shvbroker, url):
    """Check that we can login with sha1 password.

    Commonly we login with plain password in tests here but we need to check
    ability to login with SHA1 hashed password as well.
    """
    nurl = dataclasses.replace(
        url,
        login=dataclasses.replace(
            url.login,
            password="57a261a7bcb9e6cf1db80df501cdd89cee82957e",
            login_type=RpcLoginType.SHA1,
        ),
    )
    client = await SimpleClient.connect(nurl)
    assert await client.call(".app", "ping") is None
    await client.disconnect()


async def test_disconnect(client):
    """Tests that multiple calls to disconnect is not an issue."""
    await client.disconnect()


async def test_reconnect(client, shvbroker):
    """Checks that client reconnects itself if broker disconnects it.

    This uses broker's API to disconnect itself. Broker will give us a new
    client ID and that way we will identify that we successfully reconnected.
    """
    client.reconnects = 2
    info = await client.call(".broker/currentClient", "info")
    await shvbroker.get_client(info["clientId"]).disconnect()
    await asyncio.sleep(0)
    assert (await client.call(".broker/currentClient", "info"))["clientId"] != info[
        "clientId"
    ]



"""RPC Exchange usage and implementation utility."""

from __future__ import annotations

from .simplebase import SimpleBase


class RpcExchange:
    """RPC Exchange connection over :class:`SimpleBase`."""

    def __init__(
        self,
        client: SimpleBase,
        path: str,
        token: str,
    ) -> None:
        self.client: SimpleBase = client
        """The client used to access the RPC file."""
        self._path = f"{path}/{token}"
        self._wbuf: bytearray = bytearray()
        self._rbuf: bytearray = bytearray()

    async def read(self, n: int = -1) -> bytes:
        """Read bytes from the exchange stream.

        This also sends any bytes that were previously written.
        """
        return b""

    # TODO readline, readexactly, and readuntil

    def write(self, data: bytes) -> None:
        """Write bytes to the exchange stream."""

    async def drain(self) -> None:
        """Drain all written bytes to the exchange stream.

        RPC Exchange is designed to do read and write with a single operation
        where reading is required to be performed before write is allowed. This
        means that this function must buffer read data. In most cases it is
        better to use :meth:`read` directly instead of this one.
        """

    async def close(self) -> None:
        """Close the exchange stream."""
